# Copyright 2023-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""New entrypoint for charmcraft."""
from __future__ import annotations

import pathlib
import shutil
from typing import Any

import craft_application
import craft_cli
from craft_application import util
from craft_parts.plugins import plugins
from overrides import override

from charmcraft import extensions, models, preprocess, services
from charmcraft.application import commands
from charmcraft.parts import BundlePlugin, CharmPlugin, ReactivePlugin
from charmcraft.services import CharmcraftServiceFactory

GENERAL_SUMMARY = """
Charmcraft helps build, package and publish operators on Charmhub.

Together with the Python Operator Framework, charmcraft simplifies
operator development and collaboration.

See https://charmhub.io/publishing for more information.
"""

APP_METADATA = craft_application.AppMetadata(
    name="charmcraft",
    summary=GENERAL_SUMMARY,
    ProjectClass=models.CharmcraftProject,
    BuildPlannerClass=models.CharmcraftBuildPlanner,
    source_ignore_patterns=["*.charm", "charmcraft.yaml"],
)

PRIME_BEHAVIOUR_CHANGE_MESSAGE = (
    "IMPORTANT: The behaviour of the 'prime' keyword has changed in Charmcraft 3. This "
    "keyword will no longer add files that would otherwise be excluded from the "
    "charm, instead filtering existing files. Additional files may be added using the "
    "'dump' plugin.\n"
    "To include extra files, see: https://juju.is/docs/sdk/include-extra-files-in-a-charm"
)


class Charmcraft(craft_application.Application):
    """Charmcraft application definition."""

    def __init__(
        self,
        app: craft_application.AppMetadata,
        services: CharmcraftServiceFactory,
    ) -> None:
        super().__init__(app=app, services=services, extra_loggers={"charmcraft"})
        self._global_args: dict[str, Any] = {}
        self._dispatcher: craft_cli.Dispatcher | None = None
        self._cli_loggers |= {"charmcraft"}

    @property
    def command_groups(self) -> list[craft_cli.CommandGroup]:
        """Return command groups."""
        return self._command_groups

    def _project_vars(self, yaml_data: dict[str, Any]) -> dict[str, str]:
        """Return a dict with project-specific variables, for a craft_part.ProjectInfo."""
        return {"version": "unversioned"}

    def _check_deprecated(self, yaml_data: dict[str, Any]) -> None:
        """Check for deprecated fields in the yaml_data."""
        # We only need to warn people once.
        if self.is_managed():
            return
        has_primed_part = False
        if "parts" in yaml_data:
            prime_changed_extensions = {"charm", "reactive", "bundle"}
            for name, part in yaml_data["parts"].items():
                if not {name, part.get("plugin", None)} & prime_changed_extensions:
                    continue
                if "prime" in part:
                    has_primed_part = True
        if has_primed_part:
            craft_cli.emit.progress(PRIME_BEHAVIOUR_CHANGE_MESSAGE, permanent=True)

    def _extra_yaml_transform(
        self, yaml_data: dict[str, Any], *, build_on: str, build_for: str | None
    ) -> dict[str, Any]:

        # Extensions get applied on as close as possible to what the user provided.
        yaml_data = extensions.apply_extensions(self.project_dir, yaml_data.copy())

        # Preprocessing "magic" to create a fully-formed charm.
        preprocess.add_default_parts(yaml_data)
        preprocess.add_bundle_snippet(self.project_dir, yaml_data)
        preprocess.add_config(self.project_dir, yaml_data)
        preprocess.add_actions(self.project_dir, yaml_data)
        preprocess.add_metadata(self.project_dir, yaml_data)

        self._check_deprecated(yaml_data)
        return yaml_data

    def _configure_services(self, provider_name: str | None) -> None:
        super()._configure_services(provider_name)
        self.services.set_kwargs(
            "package",
            project_dir=self.project_dir,
            build_plan=self._build_plan,
        )

    def configure(self, global_args: dict[str, Any]) -> None:
        """Configure the application using any global arguments."""
        super().configure(global_args)
        self._global_args = global_args

    def _get_dispatcher(self) -> craft_cli.Dispatcher:
        """Get the dispatcher, with a charmcraft-specific side-effect of storing it on the app."""
        self._dispatcher = super()._get_dispatcher()
        return self._dispatcher

    @override
    def _get_app_plugins(self) -> dict[str, plugins.PluginType]:
        return {"charm": CharmPlugin, "bundle": BundlePlugin, "reactive": ReactivePlugin}

    @override
    def _pre_run(self, dispatcher: craft_cli.Dispatcher) -> None:
        """Override to get project_dir early."""
        super()._pre_run(dispatcher)
        if not self.is_managed() and not getattr(dispatcher.parsed_args(), "project_dir", None):
            self.project_dir = pathlib.Path().expanduser().resolve()

    def run_managed(self, platform: str | None, build_for: str | None) -> None:
        """Run charmcraft in managed mode.

        Overrides the craft-application managed mode runner to move packed files
        as needed.
        """
        dispatcher = self._dispatcher or self._get_dispatcher()
        command = dispatcher.load_command(self.app_config)
        self._work_dir = self.project_dir

        super().run_managed(platform, build_for)

        if not self.is_managed() and isinstance(command, commands.PackCommand):
            if output_dir := getattr(dispatcher.parsed_args(), "output", None):
                output_path = pathlib.Path(output_dir).resolve()
                output_path.mkdir(parents=True, exist_ok=True)
                package_file_path = self._work_dir / ".charmcraft_output_packages.txt"
                if package_file_path.exists():
                    package_files = package_file_path.read_text().splitlines(keepends=False)
                    package_file_path.unlink(missing_ok=True)
                    for filename in package_files:
                        shutil.move(str(self._work_dir / filename), output_path / filename)

    def _expand_environment(self, yaml_data: dict[str, Any], build_for: str) -> None:
        """Perform expansion of project environment variables.

        :param yaml_data: The project's yaml data.
        :param build_for: The architecture to build for.
        """
        if "-" in build_for:
            build_for = util.get_host_architecture()
            craft_cli.emit.debug(
                "Expanding environment variables with the host architecture "
                f"{build_for!r} as the build-for architecture because multiple "
                "run-on architectures were specified."
            )
        super()._expand_environment(yaml_data, build_for)


def main() -> int:
    """Run craft-application based charmcraft."""
    charmcraft_services = services.CharmcraftServiceFactory(app=APP_METADATA)

    app = Charmcraft(app=APP_METADATA, services=charmcraft_services)

    commands.fill_command_groups(app)

    return app.run()



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
import pathlib

import pytest
import yaml

from charmcraft import const
from charmcraft.errors import DuplicateCharmsError, InvalidCharmPathError
from charmcraft.utils.project import find_charm_sources, get_charm_name_from_path

BASIC_CHARM_MAP = {
    "charm1": pathlib.Path("charms/charm1/"),
    "charm2": pathlib.Path("charms/charm2/"),
    "operator1": pathlib.Path("operators/operator1"),
    "operator2": pathlib.Path("operators/operator2"),
    "basic": pathlib.Path("basic"),
    "basic2": pathlib.Path("basic2"),
}


def test_find_charm_sources_empty_directory(tmp_path):
    actual = find_charm_sources(tmp_path, ["a", "b", "c"])

    assert actual == {}


@pytest.mark.parametrize("fake_charms", [BASIC_CHARM_MAP])
def test_find_charm_sources_finds_values(tmp_path, fake_charms, build_charm_directory):
    expected = build_charm_directory(tmp_path, fake_charms)

    actual = find_charm_sources(tmp_path, fake_charms.keys())

    assert actual == expected


@pytest.mark.parametrize(
    "fake_charms",
    [{"basic": pathlib.Path("charms/basic"), "basic2": pathlib.Path("charms/basic2")}],
)
def test_find_charm_sources_with_symlinks(tmp_path, build_charm_directory, fake_charms):
    symlinks_path = tmp_path / "operators"
    symlinks_path.mkdir()
    expected = build_charm_directory(tmp_path, fake_charms)
    for name, path in fake_charms.items():
        (symlinks_path / name).symlink_to(tmp_path / path, target_is_directory=True)

    actual = find_charm_sources(tmp_path, fake_charms)

    assert actual == expected


@pytest.mark.parametrize("fake_charms", [BASIC_CHARM_MAP])
def test_find_charm_sources_extra_charms(tmp_path, build_charm_directory, fake_charms):
    expected = build_charm_directory(tmp_path, fake_charms)
    invalid_charm = tmp_path / "charms/invalid"
    build_charm_directory(tmp_path, {"invalid": invalid_charm})

    actual = find_charm_sources(tmp_path, fake_charms)

    assert actual == expected


@pytest.mark.parametrize("fake_charms", [BASIC_CHARM_MAP])
def test_find_charm_sources_non_matching_path(tmp_path, build_charm_directory, fake_charms):
    charms = {name: path.with_name(f"non_matching_{name}") for name, path in fake_charms.items()}
    build_charm_directory(tmp_path, charms)

    actual = find_charm_sources(tmp_path, fake_charms)

    assert actual == {}


def test_find_charm_sources_duplicates(check, tmp_path, build_charm_directory):
    fake_charms = {"charm1": pathlib.Path("charm1")}
    build_charm_directory(tmp_path, fake_charms)
    build_charm_directory(tmp_path, {"charm1": pathlib.Path("charms/charm1")})
    build_charm_directory(tmp_path, {"charm1": pathlib.Path("operators/charm1")})
    expected = DuplicateCharmsError(
        {
            "charm1": [
                pathlib.Path("charm1"),
                pathlib.Path("charms/charm1"),
                pathlib.Path("operators/charm1"),
            ]
        }
    )

    with pytest.raises(DuplicateCharmsError) as exc_info:
        find_charm_sources(tmp_path, fake_charms)

    check.equal(exc_info.value.args, expected.args)
    check.equal(exc_info.value.resolution, expected.resolution)


@pytest.mark.parametrize(
    ("name", "path"),
    [
        ("test1", "test1"),
        ("test1", "charms/test1"),
        ("test1", "operators/test1"),
    ],
)
def test_get_charm_name_from_path_success(tmp_path, build_charm_directory, name, path):
    build_charm_directory(tmp_path, {name: path})

    actual = get_charm_name_from_path(tmp_path / path)

    assert actual == name


@pytest.mark.parametrize(
    ("name", "path"),
    [
        ("test1", "test1"),
        ("test1", "charms/test1"),
        ("test1", "operators/test1"),
    ],
)
def test_get_charm_name_from_path_bundle(tmp_path, build_charm_directory, name, path):
    build_charm_directory(tmp_path, {name: path}, file_type="bundle")
    full_path = tmp_path / path

    with pytest.raises(InvalidCharmPathError) as exc_info:
        get_charm_name_from_path(full_path)

    assert exc_info.value.args[0] == f"Path does not contain source for a valid charm: {full_path}"


@pytest.mark.parametrize(
    ("name", "path", "del_file"),
    [
        ("test1", "test1", const.CHARMCRAFT_FILENAME),
        ("test1", "charms/test1", const.METADATA_FILENAME),
        ("test1", "operators/test1", const.CHARMCRAFT_FILENAME),
    ],
)
def test_get_charm_name_from_path_missing_file(
    tmp_path, build_charm_directory, name, path, del_file
):
    build_charm_directory(tmp_path, {name: path})
    full_path = tmp_path / path
    (full_path / del_file).unlink()

    with pytest.raises(InvalidCharmPathError) as exc_info:
        get_charm_name_from_path(full_path)

    assert exc_info.value.args[0] == f"Path does not contain source for a valid charm: {full_path}"


@pytest.mark.parametrize(
    ("name", "path"),
    [
        ("test1", "test1"),
        ("test1", "charms/test1"),
        ("test1", "operators/test1"),
    ],
)
def test_get_charm_name_from_path_wrong_name(tmp_path, build_charm_directory, name, path):
    build_charm_directory(tmp_path, {name: path}, file_type="bundle")
    full_path = tmp_path / path
    with (full_path / const.METADATA_FILENAME).open("w") as file:
        yaml.safe_dump({"naam": "not a name"}, file)

    with pytest.raises(InvalidCharmPathError) as exc_info:
        get_charm_name_from_path(full_path)

    assert exc_info.value.args[0] == f"Path does not contain source for a valid charm: {full_path}"



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for package service."""
import contextlib
import datetime
import pathlib

import freezegun
import pytest
import pytest_check

import charmcraft
from charmcraft import const, models, services
from charmcraft.application.main import APP_METADATA


@pytest.fixture
def package_service(fake_path, service_factory, default_build_plan):
    fake_project_dir = fake_path
    svc = services.PackageService(
        app=APP_METADATA,
        project=service_factory.project,
        services=service_factory,
        project_dir=fake_project_dir,
        build_plan=default_build_plan,
    )
    service_factory.package = svc
    return svc


@pytest.mark.parametrize(
    "project_path",
    [
        pytest.param(path, id=path.name)
        for path in (pathlib.Path(__file__).parent / "sample_projects").iterdir()
    ],
)
@freezegun.freeze_time(datetime.datetime(2020, 3, 14, 0, 0, 0, tzinfo=datetime.timezone.utc))
def test_write_metadata(monkeypatch, fs, package_service, project_path):
    monkeypatch.setattr(charmcraft, "__version__", "3.0-test-version")
    with contextlib.suppress(FileExistsError):
        fs.add_real_directory(project_path)
    test_prime_dir = pathlib.Path("/prime")
    fs.create_dir(test_prime_dir)
    expected_prime_dir = project_path / "prime"

    project = models.CharmcraftProject.from_yaml_file(project_path / "project" / "charmcraft.yaml")
    project._started_at = datetime.datetime.now(tz=datetime.timezone.utc)
    package_service._project = project

    package_service.write_metadata(test_prime_dir)

    for file in expected_prime_dir.iterdir():
        pytest_check.equal((test_prime_dir / file.name).read_text(), file.read_text())


@pytest.mark.parametrize(
    "project_path",
    [
        pytest.param(path, id=path.name)
        for path in (pathlib.Path(__file__).parent / "sample_projects").iterdir()
    ],
)
@freezegun.freeze_time(datetime.datetime(2020, 3, 14, 0, 0, 0, tzinfo=datetime.timezone.utc))
def test_overwrite_metadata(monkeypatch, fs, package_service, project_path):
    """Test that the metadata file gets rewritten for a charm.

    Regression test for https://github.com/canonical/charmcraft/issues/1654
    """
    monkeypatch.setattr(charmcraft, "__version__", "3.0-test-version")
    with contextlib.suppress(FileExistsError):
        fs.add_real_directory(project_path)
    test_prime_dir = pathlib.Path("/prime")
    fs.create_dir(test_prime_dir)
    expected_prime_dir = project_path / "prime"

    project = models.CharmcraftProject.from_yaml_file(project_path / "project" / "charmcraft.yaml")
    project._started_at = datetime.datetime.now(tz=datetime.timezone.utc)
    package_service._project = project

    fs.create_file(test_prime_dir / const.METADATA_FILENAME, contents="INVALID!!")

    package_service.write_metadata(test_prime_dir)

    for file in expected_prime_dir.iterdir():
        pytest_check.equal((test_prime_dir / file.name).read_text(), file.read_text())


@freezegun.freeze_time(datetime.datetime(2020, 3, 14, 0, 0, 0, tzinfo=datetime.timezone.utc))
def test_no_overwrite_reactive_metadata(monkeypatch, fs, package_service):
    """Test that the metadata file doesn't get overwritten for a reactive charm..

    Regression test for https://github.com/canonical/charmcraft/issues/1654
    """
    monkeypatch.setattr(charmcraft, "__version__", "3.0-test-version")
    project_path = pathlib.Path(__file__).parent / "sample_projects" / "basic-reactive"
    with contextlib.suppress(FileExistsError):
        fs.add_real_directory(project_path)
    test_prime_dir = pathlib.Path("/prime")
    fs.create_dir(test_prime_dir)
    test_stage_dir = pathlib.Path("/stage")
    fs.create_dir(test_stage_dir)
    fs.create_file(test_stage_dir / const.METADATA_FILENAME, contents="INVALID!!")

    project = models.CharmcraftProject.from_yaml_file(project_path / "project" / "charmcraft.yaml")
    project._started_at = datetime.datetime.now(tz=datetime.timezone.utc)
    package_service._project = project

    package_service.write_metadata(test_prime_dir)

    assert not (test_prime_dir / const.METADATA_FILENAME).exists()



# Copyright 2023-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Project-related models for Charmcraft."""

import abc
import datetime
import pathlib
import re
import textwrap
from collections.abc import Iterable, Iterator
from typing import (
    Annotated,
    Any,
    Literal,
    cast,
)

import pydantic
import pydantic.v1
from craft_application import errors, models, util
from craft_application.util import safe_yaml_load
from craft_cli import CraftError
from craft_providers import bases
from pydantic import dataclasses
from typing_extensions import Self

from charmcraft import const, preprocess, utils
from charmcraft.const import (
    BaseStr,
    BuildBaseStr,
)
from charmcraft.models import charmcraft
from charmcraft.models.charmcraft import (
    AnalysisConfig,
    BasesConfiguration,
    Charmhub,
    Links,
)
from charmcraft.parts import process_part_config

CharmcraftSummaryStr = Annotated[
    str,
    models.SummaryStr,
    pydantic.StringConstraints(max_length=200),
    # Maximum length was set to 200 characters because the 78 character maximum
    # inherited from craft-application is too restrictive, as several hundred charms
    # already exceed this maximum.
    # Eventually this limit will be reduced, ideally to 78 characters, though that may
    # never happen entirely. Reductions will only occur on major releases.
    # https://github.com/canonical/charmcraft/issues/1598
]


def get_charm_file_platform_str(bases: Iterable[charmcraft.Base]) -> str:
    """Get the "platform" section of a charm file name from an iterable of bases."""
    base_strings = []
    for base in bases:
        name = base.name
        version = base.channel
        architectures = "-".join(base.architectures)
        base_strings.append(f"{name}-{version}-{architectures}")
    return "_".join(base_strings)


CharmPlatform = Annotated[str, pydantic.StringConstraints(min_length=4, strict=True)]


class CharmLib(models.CraftBaseModel):
    """A Charm library dependency for this charm."""

    lib: str = pydantic.Field(
        title="Library Path (e.g. my-charm.my_library)",
        pattern=r"[a-z][a-z0-9_-]+\.[a-z][a-z0-9_]+",
    )
    version: str = pydantic.Field(
        title="Version filter for the charm. Either an API version or a specific [api].[patch].",
        pattern=r"[0-9]+(\.[0-9]+)?",
    )

    @pydantic.field_validator("lib", mode="before")
    def _validate_name(cls, value: str) -> str:
        """Validate the lib field, providing a useful error message on failure."""
        charm_name, _, lib_name = str(value).partition(".")
        if not charm_name or not lib_name:
            raise ValueError(
                f"Library name invalid. Expected '[charm_name].[lib_name]', got {value!r}"
            )
        # Accept python-importable charm names, but convert them to store-accepted names.
        if "_" in charm_name:
            charm_name = charm_name.replace("_", "-")
        if not re.fullmatch("[a-z0-9_-]+", charm_name):
            raise ValueError(
                f"Invalid charm name for lib {value!r}. Value {charm_name!r} is invalid."
            )
        if not re.fullmatch("[a-z0-9_]+", lib_name):
            raise ValueError(
                f"Library name {lib_name!r} is invalid. Library names must be valid Python module names."
            )
        return f"{charm_name}.{lib_name}"

    @pydantic.field_validator("version", mode="before")
    def _validate_api_version(cls, value: str) -> str:
        """Validate the API version field, providing a useful error message on failure."""
        api, *_ = str(value).partition(".")
        try:
            int(api)
        except ValueError:
            raise ValueError(f"API version not valid. Expected an integer, got {api!r}") from None
        return str(value)

    @pydantic.field_validator("version", mode="before")
    def _validate_patch_version(cls, value: str) -> str:
        """Validate the optional patch version, providing a useful error message."""
        api, separator, patch = value.partition(".")
        if not separator:
            return value
        try:
            int(patch)
        except ValueError:
            raise ValueError(
                f"Patch version not valid. Expected an integer, got {patch!r}"
            ) from None
        return value

    @property
    def api_version(self) -> int:
        """The API version needed for this library."""
        return int(self.version.partition(".")[0])

    @property
    def patch_version(self) -> int | None:
        """The patch version needed for this library, or None if no patch version is specified."""
        api, _, patch = self.version.partition(".")
        if not patch:
            return None
        return int(patch)


@dataclasses.dataclass
class CharmBuildInfo(models.BuildInfo):
    """Information about a single build option, with charmcraft-specific info.

    With CharmBuildInfo, the build_for may also be the string "multi", meaning the charm
    is expected to run on multiple architectures.
    """

    platform: CharmPlatform

    build_for_bases: list[charmcraft.Base]
    """Charmcraft base to build for, including potentially multiple architectures."""
    bases_index: int
    """Index of the base configuration in charmcraft.yaml."""
    build_on_index: int
    """Index of this base configuration's build-on option."""

    @classmethod
    def from_build_on_run_on(
        cls: type[Self],
        build_on_base: charmcraft.Base,
        build_on_arch: str,
        run_on: list[charmcraft.Base],
        *,
        bases_index: int,
        build_on_index: int,
    ) -> Self:
        """Create a single CharmBuildInfo from a build_on base and run_on bases.

        :param build_on_base: A Base object defining a base on which to build
        :param build_on_arch: The architecture on which to run the build (e.g. "amd64")
        :param run_on: A list of bases which this charm should run on after the build.
        :param bases_index: The index of the BasesConfiguration
        :param build_on_index: Which build-on value from the BasesConfiguration to use
        """
        base = bases.BaseName(name=build_on_base.name, version=build_on_base.channel)

        all_architectures = set()
        for run_on_base in run_on:
            all_architectures.update(run_on_base.architectures)

        build_for = "-".join(sorted(all_architectures))

        platform = get_charm_file_platform_str(run_on)

        return cls(
            platform=platform,
            build_on=build_on_arch,
            build_for=build_for,
            base=base,
            build_for_bases=run_on,
            bases_index=bases_index,
            build_on_index=build_on_index,
        )

    @classmethod
    def gen_from_bases_configurations(
        cls: type[Self], *bases_configs: charmcraft.BasesConfiguration
    ) -> Iterator[Self]:
        """Generate CharmBuildInfo objects from a BasesConfiguration object.

        :param bases_config: One or more BasesConfiguration objects from which to generate
            CharmBuildInfo objects.
        :returns: A list of CharmBuildInfo objects from this BasesConfiguration.

        Example 1: a simple charm:
            bases:
              - name: ubuntu
                channel: 24.04
        This gets expanded to a standard long-form:
            bases:
              - build-on:
                  - name: ubuntu
                    channel: "24.04"
                run-on:
                  - name: ubuntu
                    channel: "24.04"
        Presuming charmcraft is run on riscv64 (if architectures are not specified charmcraft will
        use the host architecture as the only architecture), it will output a list containing the
        following single BuildInfo object:
            CharmBuildInfo(
                build_on="riscv64",
                build_for="riscv64",
                base=BaseName(name="ubuntu", channel="24.04"),
                build_for_base=BaseName(name="ubuntu", channel="24.04"),
                bases_index=0,
                build_on_index=0
            )

        Example 2: a more complex set of bases:
            bases:
              - build-on:
                  - name: ubuntu
                    channel: "24.04"
                    architectures: ["amd64", "riscv64"]
                  - name: ubuntu
                    channel: "22.04"
                    architectures: ["arm64"]
                run-on:
                  - name: ubuntu
                    channel: "22.04"
                    architectures: ["amd64", "arm64"]
                  - name: ubuntu
                    channel: "24.04"
                    architectures: ["amd64", "arm64", "riscv64"]
        This will result in the following builds in the plan:
        [
            CharmBuildInfo(
                build_on="amd64",
                build_for="multi",
                base=BaseName(name="ubuntu", channel="24.04"),
                build_for_bases=[
                    Base(name="ubuntu", channel="22.04", architectures=["amd64", "arm64"]),
                    Base(name="ubuntu", channel="24.04", architectures=["amd64", "arm64", "riscv64"])
                ]
                bases_index=0,
                build_on_index=0
            ),
            CharmBuildInfo(
                build_on="riscv64",
                build_for="multi",
                base=BaseName(name="ubuntu", channel="24.04"),
                build_for_bases=[
                    Base(name="ubuntu", channel="22.04", architectures=["amd64", "arm64"]),
                    Base(name="ubuntu", channel="24.04", architectures=["amd64", "arm64", "riscv64"])
                ]
                bases_index=0,
                build_on_index=0
            ),
            CharmBuildInfo(
                build_on="arm64",
                build_for="multi",
                base=BaseName(name="ubuntu", channel="22.04"),
                build_for_bases=[
                    Base(name="ubuntu", channel="22.04", architectures=["amd64", "arm64"]),
                    Base(name="ubuntu", channel="24.04", architectures=["amd64", "arm64", "riscv64"])
                ]
                bases_index=0,
                build_on_index=1
            ),
        ]

        Here the string "multi" defines a destination platform that has multiple architectures.
        """
        for bases_index, bases_config in enumerate(bases_configs):
            for build_on_index, build_on_base in enumerate(bases_config.build_on):
                for build_on_arch in build_on_base.architectures:
                    yield cls.from_build_on_run_on(
                        build_on_base,
                        build_on_arch,
                        bases_config.run_on,
                        bases_index=bases_index,
                        build_on_index=build_on_index,
                    )


class CharmcraftBuildPlanner(models.BuildPlanner):
    """Build planner for Charmcraft."""

    type: str = ""
    bases: list[BasesConfiguration] = pydantic.Field(default_factory=list)
    base: str | None = None
    build_base: str | None = None
    platforms: dict[str, models.Platform | None] | None = None  # type: ignore[assignment]

    def get_build_plan(self) -> list[models.BuildInfo]:
        """Get build bases for this charm.

        This method provides a flattened version of every way to build the charm, unfiltered.

        If a charm uses the older "bases" model, it defers to
        `CharmBuildInfo.gen_from_bases_configurations'. Otherwise, it generates the BuildInfo
        as expected with platforms.
        """
        if self.type == "bundle":
            # A bundle can build anywhere, so just present the current system.
            current_arch = util.get_host_architecture()
            current_base = utils.get_os_platform()
            return [
                models.BuildInfo(
                    platform=current_arch,
                    build_on=current_arch,
                    build_for=current_arch,
                    base=bases.BaseName(name=current_base.system, version=current_base.release),
                )
            ]
        if not self.base:
            return list(CharmBuildInfo.gen_from_bases_configurations(*self.bases))

        build_base = self.build_base or self.base
        base_name, _, base_version = build_base.partition("@")
        base = bases.BaseName(name=base_name, version=base_version)

        if self.platforms is None:
            raise CraftError("Must define at least one platform.")
        build_infos = []
        for platform_name, platform in self.platforms.items():
            if platform is None:
                if platform_name not in const.SUPPORTED_ARCHITECTURES:
                    raise CraftError(
                        f"Invalid platform {platform_name}.",
                        details="A platform name must either be a valid architecture name or the "
                        "platform must specify one or more build-on and build-for architectures.",
                    )
                build_infos.append(
                    models.BuildInfo(
                        platform_name,
                        build_on=platform_name,
                        build_for=platform_name,
                        base=base,
                    )
                )
            else:
                # TODO: this should go to craft-platforms, so silence mypy for now.
                for build_on in platform.build_on:  # type: ignore[union-attr]
                    build_infos.extend(
                        [
                            models.BuildInfo(
                                platform_name,
                                build_on=str(build_on),
                                build_for=str(build_for),
                                base=base,
                            )
                            for build_for in platform.build_for  # type: ignore[union-attr]
                        ]
                    )
        return build_infos


class CharmcraftProject(models.Project, metaclass=abc.ABCMeta):
    """A craft-application compatible version of a Charmcraft project.

    This is a Project definition for charmcraft commands that are run through
    craft-application rather than the legacy charmcraft entrypoint. Eventually
    it will be the only form of the project.

    This inherits from CraftBaseModel rather than from the base craft-application Project
    in order to preserve field order. It's registered as a virtual child class below.
    """

    type: Literal["charm", "bundle"]
    title: models.ProjectTitle | None = None
    summary: CharmcraftSummaryStr | None = None
    description: str | None = None

    analysis: AnalysisConfig | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            How analysis done on the charm will behave.

            Currently the only options are to ignore attributes or linters."""
        ),
    )
    charmhub: Charmhub | None = pydantic.Field(
        default=None, description="(DEPRECATED): Configuration for accessing charmhub."
    )
    parts: dict[str, dict[str, Any]] = pydantic.Field(default_factory=dict)

    # Default project properties that Charmcraft currently does not use. Types are set
    # to be Optional[None], preventing them from being used, but allow them to be used
    # by the application.
    version: Literal["unversioned"] = "unversioned"  # type: ignore[assignment]
    license: None = None
    # These are inside the "links" child model.
    contact: None = None
    issues: None = None
    source_code: None = None
    charm_libs: list[CharmLib] = pydantic.Field(
        default_factory=list, title="List of libraries to use for this charm"
    )

    # These private attributes are not part of the project model but are attached here
    # because Charmcraft uses this metadata.
    _started_at: datetime.datetime = pydantic.PrivateAttr(
        default_factory=lambda: datetime.datetime.now(tz=datetime.timezone.utc)
    )
    _valid: bool = pydantic.PrivateAttr(default=False)

    @property
    def started_at(self) -> datetime.datetime:
        """Get the time that Charmcraft started running."""
        return self._started_at

    @classmethod
    def unmarshal(cls, data: dict[str, Any]):
        """Create a Charmcraft project from a dictionary of data."""
        if cls is not CharmcraftProject:
            return cls.model_validate(data)
        project_type = data.get("type")
        if project_type == "charm":
            if "bases" in data:
                return BasesCharm.unmarshal(data)
            return PlatformCharm.unmarshal(data)
        if project_type == "bundle":
            return Bundle.unmarshal(data)
        raise ValueError(f"field type cannot be {project_type!r}")

    @classmethod
    def from_yaml_file(cls, path: pathlib.Path) -> Self:
        """Instantiate this model from a YAML file.

        For use with craft-application.
        """
        try:
            with path.open() as file:
                data = safe_yaml_load(file)
        except FileNotFoundError:
            raise CraftError(f"Could not find charmcraft.yaml at '{path}'")
        except OSError as exc:
            raise CraftError(
                f"Error parsing charmcraft.yaml at '{path}'", details=exc.strerror
            ) from exc

        if not isinstance(data, dict):
            raise errors.CraftValidationError(
                "Invalid 'charmcraft.yaml' file",
                details=f"File generated a {type(data)} object, expected a dictionary",
                resolution="Ensure 'charmcraft.yaml' is valid",
                reportable=False,
                docs_url="https://juju.is/docs/sdk/charmcraft-yaml",
            )

        project_dir = path.parent

        preprocess.add_default_parts(data)
        preprocess.add_bundle_snippet(project_dir, data)
        preprocess.add_metadata(project_dir, data)
        preprocess.add_config(project_dir, data)
        preprocess.add_actions(project_dir, data)

        try:
            project = cls.unmarshal(data)
        except pydantic.ValidationError as err:
            raise errors.CraftValidationError.from_pydantic(err, file_name=path.name)
        except ValueError as err:
            error_str = "\n".join(f"- {arg}" for arg in err.args)
            raise errors.CraftValidationError(
                f"Bad charmcraft.yaml content:\n{error_str}",
            )

        return project

    @pydantic.model_validator(mode="before")
    @classmethod
    def _preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Preprocess any values that charmcraft infers, before attribute validation."""
        if "type" not in values:
            raise ValueError("Project type must be declared in charmcraft.yaml.")

        return values

    @pydantic.field_validator("parts", mode="before")
    @classmethod
    def _preprocess_parts(
        cls, parts: dict[str, dict[str, Any]] | None, info: pydantic.ValidationInfo
    ) -> dict[str, dict[str, Any]]:
        """Preprocess parts object for a charm or bundle, creating an implicit part if needed."""
        if parts is not None and not isinstance(parts, dict):
            raise TypeError("'parts' in charmcraft.yaml must conform to the charmcraft.yaml spec.")
        if not parts:
            if info.config and info.config.get("title") == "Bundle":
                parts = {"bundle": {"plugin": "bundle"}}
            elif "type" in info.data:
                parts = {info.data["type"]: {"plugin": info.data["type"]}}
            else:
                parts = {}
        for name, part in parts.items():
            if not isinstance(part, dict):
                raise TypeError(f"part {name!r} must be a dictionary.")
            # implicit plugin fixup
            if "plugin" not in part:
                part["plugin"] = name

        for name, part in parts.items():
            if name == "charm" and part["plugin"] == "charm":
                part.setdefault("source", ".")

            if name == "bundle" and part["plugin"] == "bundle":
                part.setdefault("source", ".")
        return {name: process_part_config(part) for name, part in parts.items()}


class CharmProject(CharmcraftProject):
    """A base class for all charm types."""

    type: Literal["charm"]
    """The type of project. Must be the string ``charm``."""
    name: models.ProjectName = pydantic.Field(
        description=textwrap.dedent(
            """\
            The name of the project on Charmhub.

            This value will be used both in the URL of the charm on Charmhub and when
            deploying the charm with juju.

            Charms should follow the
            `charm naming guidelines <https://juju.is/docs/sdk/naming>`_."""
        ),
        examples=[
            "mysql",
            "mysql-k8s",
        ],
    )
    summary: CharmcraftSummaryStr = pydantic.Field(  # pyright: ignore[reportGeneralTypeIssues]
        description="A brief (one-line) summary of your charm.",
    )
    description: str = pydantic.Field(  # pyright: ignore[reportGeneralTypeIssues]
        description="A multi-line summary of your charm."
    )

    parts: dict[str, dict[str, Any]] = pydantic.Field(
        default={"charm": {"plugin": "charm", "source": "."}},
        description=textwrap.dedent(
            """\
            Configures the various mechanisms to obtain, process and prepare data from
            different sources that end up being a part of the final charm.

            Keys are user-defined part names. The value of each key is a map where keys
            are part names. Charmcraft provides 3 plugins: charm, bundle, reactive.

            Example::

                parts:
                  libs:
                    plugin: dump
                    source: /usr/local/lib/
                    organize:
                      "libxxx.so*": lib/
                    prime:
                      - lib/""",
        ),
    )

    actions: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            Defines one or more actions.

            This key is equivalent to the
            `actions.yaml file <https://juju.is/docs/sdk/actions-yaml>`_."""
        ),
    )
    assumes: list[str | dict[str, list | dict]] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            Explicitly state features a Juju model must be able to provide for a
            successful deployment of this charm. When a charm includes such
            requirements, Juju performs a pre-deployment check and displays
            user-friendly error messages if a feature requirement cannot be met by the
            model that the user is trying to deploy the charm to. If the assumes
            section of the charm metadata is omitted, Juju will make a best-effort
            attempt to deploy the charm, and users must rely on the output of
            ``juju status`` to figure out whether the deployment was successful.

            The key consists of a list of features that can be given either directly
            or, depending on the complexity of the condition you want to enforce,
            nested under one or both of the boolean expressions any-of or all-of,
            as shown below. In order for a charm to be deployed, all entries in the
            assumes block must be satisfied.

            Structure::

                assumes:
                  - <feature-1>
                  - any-of:
                    - <feature-2>
                    - <feature-3>
                  - all-of:
                    - <feature-4>
                    - <feature-5>

            Juju version requirements can be specified with a string such as
            ``juju >= 3.5`` or ``juju < 4.0``. A full list of supported features
            can be found in the
            `Juju documentation <https://juju.is/docs/juju/supported-features>`_.
            """
        ),
    )
    charm_user: Literal["root", "sudoer", "non-root"] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            Specifies that the charm code does not need to be run as root. Possible
            values are:

            - ``root``: the charm will run as root
            - ``sudoer``: the charm will run as a non-root user with access to root
              privileges using ``sudo``.
            - ``non-root``: the charm will run as a non-root user without ``sudo``.

            Only affects Kubernetes charms on Juju 3.6.0 or later. If not specified,
            Juju will use
            `its default behaviour <https://juju.is/docs/sdk/metadata-yaml#heading--charm-user>`_.
            """
        ),
    )
    containers: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            Define a map of containers to be created adjacent to the charm (as a
            sidecar, in the same pod).

            This is required for Kubernetes charms.

            This key consists of a dictionary mapping container names to their
            specifications. Each container can be specified in terms of ``resource``,
            ``bases`` and ``mounts``, where one of either the ``resource`` or the
            ``bases`` subkeys must be defined and ``mounts`` is optional.

            - ``resource`` is the name of an OCI image resource used to create the
              container (that you will then define further in the resources block).
            - ``bases`` is a list of bases to be used for resolving a container image,
              in descending order of preference. To use it, specify a base name (for
              example, ``ubuntu`` or ``centos``), a ``channel`` and an
              ``architecture``.
            - ``mounts`` is a list of mounted storage volumes for this container. To
              use it, specify the name of the storage to mount from the charm
              storage and, optionally, the location where to mount the storage.

            Structure::

                containers:
                  <container name>:
                    resource: <resource name>
                    bases:
                      - name: <base name>
                        channel: <track[/risk][/branch]>
                        architectures:
                          - <architecture>
                    mounts:
                      - storage: <storage name>
                        location: <path>"""
        ),
        examples=[
            {
                "super-app": {
                    "resource": "super-app-image",
                    "mounts": [{"storage": "logs", "location": "/logs"}],
                }
            }
        ],
    )
    devices: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            Devices the charm needs.

            Structure::

                devices:
                  <device name>:
                    type: gpu | nvidia.com/gpu | amd.com/gpu
                    description: <Optional description>
                    countmin: <Optional minimum number requested>
                    countmax: <Optional maximum number requested>"""
        ),
        examples=[
            {
                "amd-gpu": {
                    "type": "amd.com/gpu",
                    "description": "Some sweet AMD GPU",
                    "countmin": 1,
                    "countmax": 100,
                },
                "nvidia-gpu": {
                    "type": "nvidia.com/gpu",
                    "description": "Some NVIDIA GPUs",
                    "countmin": 20,
                },
            },
            {
                "gpus": {
                    "type": "gpu",
                    "description": "A bunch of GPUs",
                    "countmin": 2,
                    "countmax": 40,
                }
            },
        ],
    )
    extra_bindings: dict[str, Any] | None = pydantic.Field(
        default=None,
        description="A key-only mapping representing extra bindings needed.",
    )
    peers: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            A map of peer relations.

            Structure::

                peers:
                  <endpoint name>:
                    interface: <Required interface name>
                    limit: <Optional: maximum number of supported connections
                    optional: <Informational only - whether the relation is required.>
                    scope: <"global" or "container" - the relation scope.>

            For more information, see
            `the Juju documentation <https://juju.is/docs/juju/relation>`_."""
        ),
        examples=[
            {
                "friend": {
                    "interface": "life",
                    "limit": 150,
                    "optional": True,
                    "scope": "container",
                }
            }
        ],
    )
    provides: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            A map of interfaces this charm provides.

            Structure::

                provides:
                  <endpoint name>:
                    interface: <Required interface name>
                    limit: <Optional: maximum number of supported connections
                    optional: <Informational only - whether the relation is required.>
                    scope: <"global" or "container" - the relation scope.>

            For more information, see
            `the Juju documentation <https://juju.is/docs/juju/relation>`_."""
        ),
        examples=[{"self": {"interface": "identity"}}],
    )
    requires: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            A map of relations this charm requires.

            Structure::

                requires:
                  <endpoint name>:
                    interface: <Required interface name>
                    limit: <Optional: maximum number of supported connections
                    optional: <Informational only - whether the relation is required.>
                    scope: <"global" or "container" - the relation scope.>

            For more information, see
            `the Juju documentation <https://juju.is/docs/juju/relation>`_."""
        ),
        examples=[
            {
                "parent": {
                    "interface": "birth",
                    "limit": 2,
                    "optional": False,
                    "scope": "global",
                }
            }
        ],
    )
    resources: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            A mapping of resources that accompany the charm.

            See first: `Juju | Charm resource <https://juju.is/docs/juju/charm-resource>`_

            Each resource is made available when the charm is deployed. NOTE:
            Kubernetes charms must declare an ``oci-image`` type resource for each
            container declared in ``containers``.

            Structure::

                # (Optional) Additional resources that accompany the charm
                resources:
                  <resource name>:
                    # (Required) The type of the resource
                    type: file | oci-image

                    # (Optional) Description of the resource and its purpose
                    description: <description>

                    # (Required: when type:file) The filename of the resource as it
                    # should appear in the filesystem.
                    filename: <filename>"""
        ),
        examples=[
            {"water": {"type": "file", "filename": "/dev/h2o"}},
            {"super-app-image": {"type": "oci-image", "description": "My app!"}},
        ],
    )
    storage: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            Storage devices requested by the charm.

            Structure::

                storage:
                  # Each key represents the name of the storage
                  <storage name>:

                    # (Required) Type of the requested storage
                    type: filesystem | block

                    # (Optional) Description of the storage requested
                    description: <description>

                    # (Optional) The mount location for filesystem stores. For
                    # multi-stores the location acts as the parent directory for each
                    # mounted store.
                    location: <location>

                    # Indicates if all units of the application share the storage.
                    # Defaults to false
                    shared: true | false

                    # Indicates if the storage should be made read-only (where
                    # possible). Defaults to false
                    read-only: true | false

                    # (Optional) The number of storage instances to be requested
                    multiple:
                      range: <n> | <n>-<m> | <n>- | <n>+

                    # (Optional) Minimum size of requested storage in forms G, GiB, GB.
                    # Size multipliers are M, G, T, P, E, Z or Y. With no multiplier
                    # supplied, M is implied.
                    minimum-size: <n> | <n><multiplier>

                    # (Optional) List of properties, only supported value is "transient"
                    properties:
                      - transient
            """
        ),
        examples=[
            {
                "jbod": {
                    "type": "block",
                    "description": "A block storage to use as swap space",
                    "shared": False,
                    "properties": ["transient"],
                },
            },
        ],
    )
    subordinate: bool | None = pydantic.Field(
        default=None,
        description="Optional boolean to declare the charm subordinate.",
        examples=[True],
    )
    terms: list[str] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            A list of terms to which the user agree by using the charm.
            These terms are not enforced by the charm, Juju or Canonical."""
        ),
        examples=[
            "Post cat pictures on Mastodon",
            "Tag your cat pictures with #caturday",
        ],
    )
    links: Links | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            (Recommended) Links to various additional information used by Charmhub."""
        ),
        examples=[
            {
                "contact": "Please send your answer to Old Pink, care of the Funny Farm, Chalfont",
                "documentation": "https://discourse.charmhub.io/t/traefik-k8s-docs-index/10778",
                "issues": "https://github.com/canonical/traefik-k8s-operator/issues",
                "source": "https://github.com/canonical/traefik-k8s-operator",
                "website": "https://charmed-kubeflow.io/",
            }
        ],
    )
    config: dict[str, Any] | None = pydantic.Field(
        default=None,
        description=textwrap.dedent(
            """\
            One or more configuration options for your charm.

            Structure::

                config:
                  options:
                    # Each option name is the name by which the charm will query the option.
                    <option name>:
                      # (Required) The type of the option
                      type: string | int | float | boolean | secret
                      # (Optional) The default value of the option
                      default: <a reasonable default value of the same type as the option>
                      # (Optional): A string describing the option. Also appears on charmhub.io
                      description: <description string>"""
        ),
        examples=[
            {
                "options": {
                    "name": {
                        "default": "Wiki",
                        "description": "The name or title of the Wiki",
                        "type": "string",
                    },
                    "skin": {
                        "default": "vector",
                        "description": "Skin to use for the wiki",
                        "type": "string",
                    },
                },
            },
        ],
    )


def _check_base_is_legacy(base: charmcraft.BaseDict) -> bool:
    """Check that the given base is a legacy base, usable with 'bases'."""
    # This pyright ignore can go away once we're on Python minimum version 3.11.
    # At that point we can mark items as required or not required.
    # https://docs.python.org/3/library/typing.html#typing.Required
    if (
        base["name"] == "ubuntu"  # pyright: ignore[reportTypedDictNotRequiredAccess]
        and base["channel"] < "24.04"  # pyright: ignore[reportTypedDictNotRequiredAccess]
    ):
        return True
    return base in ({"name": "centos", "channel": "7"}, {"name": "almalinux", "channel": "9"})


def _validate_base(
    base: charmcraft.BaseDict | charmcraft.LongFormBasesDict,
) -> charmcraft.LongFormBasesDict:
    if "name" in base:  # Convert short form to long form
        base = cast(charmcraft.LongFormBasesDict, {"build-on": [base], "run-on": [base]})
    else:  # Cast to long form since we know it is one.
        base = cast(charmcraft.LongFormBasesDict, base)

    # Ensure we're only allowing legacy bases.
    for build_base in base["build-on"]:
        if not _check_base_is_legacy(build_base):
            raise ValueError(f"Base requires 'platforms' definition: {build_base}")
    for run_base in base["run-on"]:
        if not _check_base_is_legacy(run_base):
            raise ValueError(f"Base requires 'platforms' definition: {run_base}")
    return base


class BasesCharm(CharmProject):
    """A charm using the deprecated ``bases`` keyword.

    This type of charm only supports the following bases:
        - Ubuntu 18.04
        - Ubuntu 20.04
        - Ubuntu 22.04
        - CentOS 7
        - Alma Linux 9
    """

    platforms: None = None  # type: ignore[assignment]

    # This is defined this way because using conlist makes mypy sad and using
    # a ConstrainedList child class has pydantic issues. This appears to be
    # solved with Pydantic 2.
    bases: list[Annotated[BasesConfiguration, pydantic.BeforeValidator(_validate_base)]] = (
        pydantic.Field(min_length=1)
    )

    base: None = None


class PlatformCharm(CharmProject):
    """Model for defining a charm using Platforms."""

    # Silencing pyright because it complains about missing default value
    base: BaseStr  # pyright: ignore[reportGeneralTypeIssues]
    build_base: BuildBaseStr | None = None
    platforms: dict[str, models.Platform | None]  # type: ignore[assignment]

    @pydantic.model_validator(mode="after")
    def _validate_dev_base_needs_build_base(self) -> Self:
        if not self.build_base and self.base in const.DEVEL_BASE_STRINGS:
            raise ValueError(
                f"Base {self.base} requires a build-base (recommended: 'build-base: ubuntu@devel')"
            )
        return self


Charm = BasesCharm | PlatformCharm


class Bundle(CharmcraftProject):
    """Model for defining a bundle."""

    type: Literal["bundle"]
    bundle: dict[str, Any] = {}
    name: models.ProjectName | None = None  # type: ignore[assignment]
    title: models.ProjectTitle | None = None
    summary: CharmcraftSummaryStr | None = None
    description: pydantic.StrictStr | None = None
    platforms: None = None  # type: ignore[assignment]

    @pydantic.model_validator(mode="before")
    @classmethod
    def preprocess_bundle(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Preprocess any values that charmcraft infers, before attribute validation."""
        if "name" not in values:
            values["name"] = values.get("bundle", {}).get("name")

        return values



# Copyright 2021-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Module to work with OCI registries."""

import base64
import gzip
import hashlib
import io
import json
import os
import tarfile
import tempfile
from typing import Any
from urllib.request import parse_http_list, parse_keqv_list

import requests
import requests_unixsocket  # type: ignore[import-untyped]
from craft_cli import CraftError, emit

# some mimetypes
CONFIG_MIMETYPE = "application/vnd.docker.container.image.v1+json"
MANIFEST_V2_MIMETYPE = "application/vnd.docker.distribution.manifest.v2+json"
LAYER_MIMETYPE = "application/vnd.docker.image.rootfs.diff.tar.gzip"
JSON_RELATED_MIMETYPES = {
    "application/json",
    "application/vnd.docker.distribution.manifest.v1+prettyjws",  # signed manifest
    MANIFEST_V2_MIMETYPE,
}
OCTET_STREAM_MIMETYPE = "application/octet-stream"

# downloads and uploads happen in chunks; this size is mostly driven by the usage in the upload
# blob, where the cost in time is similar for small and large chunks (we need to balance having
# it large enough for speed, but not too large because of memory consumption)
CHUNK_SIZE = 2**20


def assert_response_ok(
    response: requests.Response, expected_status: int = 200
) -> dict[str, Any] | None:
    """Assert the response is ok."""
    if response.status_code != expected_status:
        ct = response.headers.get("Content-Type", "")
        if ct.split(";")[0] in JSON_RELATED_MIMETYPES:
            errors = response.json().get("errors")
        else:
            errors = None
        raise CraftError(
            "Wrong status code from server "
            f"(expected={expected_status}, got={response.status_code})",
            details=f"errors={errors} headers={response.headers}",
        )

    if response.headers.get("Content-Type") not in JSON_RELATED_MIMETYPES:
        return None

    result = response.json()
    if "errors" in result:
        raise CraftError("Response with errors from server: {}".format(result["errors"]))
    return result


class OCIRegistry:
    """Interface to a generic OCI Registry."""

    def __init__(self, server, image_name, *, username="", password=""):
        self.server = server
        self.image_name = image_name
        self.auth_token = None

        if username:
            _u_p = f"{username}:{password}"
            self.auth_encoded_credentials = base64.b64encode(_u_p.encode("ascii")).decode("ascii")
        else:
            self.auth_encoded_credentials = None

    def __eq__(self, other):
        return (
            self.server == other.server
            and self.image_name == other.image_name
            and self.auth_encoded_credentials == other.auth_encoded_credentials
        )

    def _authenticate(self, auth_info):
        """Get the auth token."""
        headers = {}
        if self.auth_encoded_credentials is not None:
            headers["Authorization"] = f"Basic {self.auth_encoded_credentials}"

        emit.trace(f"Authenticating! {auth_info}")
        url = "{realm}?service={service}&scope={scope}".format_map(auth_info)
        response = requests.get(url, headers=headers)

        result = assert_response_ok(response)
        return result["token"]

    def _get_url(self, subpath):
        """Build the URL completing the subpath."""
        return f"{self.server}/v2/{self.image_name}/{subpath}"

    def _get_auth_info(self, response):
        """Parse a 401 response and get the needed auth parameters."""
        www_auth = response.headers["Www-Authenticate"]
        if not www_auth.startswith("Bearer "):
            raise ValueError("Bearer not found")
        return parse_keqv_list(parse_http_list(www_auth[7:]))

    def _hit(self, method, url, headers=None, log=True, **kwargs):
        """Hit the specific URL, taking care of the authentication."""
        if headers is None:
            headers = {}
        if self.auth_token is not None:
            headers["Authorization"] = f"Bearer {self.auth_token}"

        if log:
            emit.trace(f"Hitting the registry: {method} {url}")
        response = requests.request(method, url, headers=headers, **kwargs)
        if response.status_code == 401:
            # token expired or missing, let's get another one and retry
            try:
                auth_info = self._get_auth_info(response)
            except (ValueError, KeyError) as exc:
                raise CraftError(f"Bad 401 response: {exc}; headers: {response.headers!r}")
            self.auth_token = self._authenticate(auth_info)
            headers["Authorization"] = f"Bearer {self.auth_token}"
            response = requests.request(method, url, headers=headers, **kwargs)

        return response

    def _is_item_already_uploaded(self, url):
        """Verify if a generic item is uploaded."""
        response = self._hit("HEAD", url)

        if response.status_code == 200:
            # item is there, done!
            uploaded = True
        elif response.status_code == 404:
            # confirmed item is NOT there
            uploaded = False
        else:
            # something else is going on, log what we have and return False so at least
            # we can continue with the upload
            emit.debug(
                f"Bad response when checking for uploaded {url!r}: "
                f"{response.status_code!r} (headers={response.headers})",
            )
            uploaded = False
        return uploaded

    def is_manifest_already_uploaded(self, reference):
        """Verify if the manifest is already uploaded, using a generic reference.

        If yes, return its digest.
        """
        emit.progress("Checking if manifest is already uploaded")
        url = self._get_url(f"manifests/{reference}")
        return self._is_item_already_uploaded(url)

    def is_blob_already_uploaded(self, reference):
        """Verify if the blob is already uploaded, using a generic reference.

        If yes, return its digest.
        """
        emit.progress("Checking if the blob is already uploaded")
        url = self._get_url(f"blobs/{reference}")
        return self._is_item_already_uploaded(url)

    def upload_manifest(self, manifest_data, reference):
        """Upload a manifest."""
        url = self._get_url(f"manifests/{reference}")
        headers = {
            "Content-Type": MANIFEST_V2_MIMETYPE,
        }
        emit.progress(f"Uploading manifest with reference {reference}")
        response = self._hit("PUT", url, headers=headers, data=manifest_data.encode("utf8"))
        assert_response_ok(response, expected_status=201)
        emit.progress("Manifest uploaded OK")

    def upload_blob(self, filepath, size, digest):
        """Upload the blob from a file."""
        # get the first URL to start pushing the blob
        emit.progress("Getting URL to push the blob")
        url = self._get_url("blobs/uploads/")
        response = self._hit("POST", url)
        assert_response_ok(response, expected_status=202)
        upload_url = response.headers["Location"]
        range_from, range_to_inclusive = (int(x) for x in response.headers["Range"].split("-"))
        emit.progress(f"Got upload URL ok with range {range_from}-{range_to_inclusive}")
        if range_from != 0:
            raise CraftError(
                "Server error: bad range received", details=f"Range={response.headers['Range']!r}"
            )

        # this `range_to_inclusive` alteration is a side effect of the range being inclusive. The
        # server tells us that it already has "0-80", means that it has 81 bytes (from 0 to 80
        # inclusive), we set from_position in 81 and read from there. Going down, "0-1" would mean
        # it has bytes 0 and 1; But "0-0" is special, it's what the server returns when it does
        # not have ANY bytes at all. So we comply with Range parameter, but addressing this
        # special case; worst think it could happen is that we start from 0 when the server
        # has 1 byte already, which is not a problem.
        if range_to_inclusive == 0:
            range_to_inclusive = -1
        from_position = range_to_inclusive + 1

        # start the chunked upload
        with open(filepath, "rb") as fh:
            with emit.progress_bar("Uploading...", size) as progress:
                if from_position:
                    fh.seek(from_position)
                    progress.advance(from_position)

                while True:
                    chunk = fh.read(CHUNK_SIZE)
                    if not chunk:
                        break

                    progress.advance(len(chunk))
                    end_position = from_position + len(chunk)
                    headers = {
                        "Content-Length": str(len(chunk)),
                        "Content-Range": f"{from_position}-{end_position}",
                        "Content-Type": OCTET_STREAM_MIMETYPE,
                    }
                    response = self._hit(
                        "PATCH", upload_url, headers=headers, data=chunk, log=False
                    )
                    assert_response_ok(response, expected_status=202)

                    upload_url = response.headers["Location"]
                    from_position += len(chunk)
        headers = {
            "Content-Length": "0",
            "Connection": "close",
        }
        emit.progress("Closing the upload")
        closing_url = f"{upload_url}&digest={digest}"

        response = self._hit("PUT", closing_url, headers=headers, data="")
        assert_response_ok(response, expected_status=201)
        emit.progress("Upload finished OK")
        if response.headers["Docker-Content-Digest"] != digest:
            raise CraftError("Server error: the upload is corrupted")


class HashingTemporaryFile(io.FileIO):
    """A temporary file that keeps the hash and length of what is written."""

    def __init__(self):
        tmp_file = tempfile.NamedTemporaryFile(mode="wb", delete=False)
        self.file_handler = tmp_file.file
        super().__init__(tmp_file.name, mode="wb")
        self.total_length = 0
        self.hasher = hashlib.sha256()

    @property
    def hexdigest(self):
        """Calculate the digest."""
        return self.hasher.hexdigest()

    def write(self, data):
        """Intercept real write to feed hasher and length count."""
        self.total_length += len(data)
        self.hasher.update(data)
        super().write(data)


class LocalDockerdInterface:
    """Functionality to interact with a local Docker daemon."""

    # the address of the dockerd socket
    dockerd_socket_baseurl = "http+unix://%2Fvar%2Frun%2Fdocker.sock"

    def __init__(self):
        self.session = requests_unixsocket.Session()

    def get_image_info_from_id(self, image_id: str) -> dict | None:
        """Get the info for a specific image using its id.

        Returns None to flag that the requested id was not found for any reason.
        """
        url = self.dockerd_socket_baseurl + f"/images/{image_id}/json"
        try:
            response = self.session.get(url)
        except requests.exceptions.ConnectionError:
            emit.debug(
                "Cannot connect to /var/run/docker.sock , please ensure dockerd is running.",
            )
            return None

        if response.status_code == 200:
            # image is there, we're fine
            return response.json()

        # 404 is the standard response to "not found", if not exactly that let's log
        # for proper debugging
        if response.status_code != 404:
            emit.debug(f"Bad response when validating local image: {response.status_code}")
            return None
        return None

    def get_image_info_from_digest(self, digest: str) -> dict | None:
        """Get the info for a specific image using its digest.

        Returns None to flag that the requested digest was not found for any reason.
        """
        url = self.dockerd_socket_baseurl + "/images/json"
        try:
            response = self.session.get(url)
        except requests.exceptions.ConnectionError:
            emit.debug(
                "Cannot connect to /var/run/docker.sock , please ensure dockerd is running.",
            )
            return None

        if response.status_code != 200:
            emit.debug(f"Bad response when validating local image: {response.status_code}")
            return None

        for image_info in response.json():
            if image_info["RepoDigests"] is None:
                continue
            if any(digest in repo_digest for repo_digest in image_info["RepoDigests"]):
                return image_info
        return None

    def get_streamed_image_content(self, image_id: str) -> requests.Response:
        """Stream the content of a specific image."""
        url = self.dockerd_socket_baseurl + f"/images/{image_id}/get"
        return self.session.get(url, stream=True)


class ImageHandler:
    """Provide specific functionalities around images."""

    def __init__(self, registry):
        self.registry = registry

    def check_in_registry(self, digest: str) -> bool:
        """Verify if the image is present in the registry."""
        return self.registry.is_manifest_already_uploaded(digest)

    def _extract_file(
        self, image_tar: str, name: str, compress: bool = False
    ) -> tuple[str, int, str]:
        """Extract a file from the tar and return its info. Optionally, gzip the content."""
        emit.progress(f"Extracting file {name!r} from local tar (compress={compress})")
        src_filehandler = image_tar.extractfile(name)
        mtime = image_tar.getmember(name).mtime

        hashing_temp_file = HashingTemporaryFile()
        if compress:
            # open the gzip file using the temporary file handler; use the original name and time
            # as 'filename' and 'mtime' correspondingly as those go to the gzip headers,
            # to ensure same final hash across different runs
            dst_filehandler = gzip.GzipFile(
                fileobj=hashing_temp_file,
                mode="wb",
                filename=os.path.basename(name),
                mtime=mtime,
            )
        else:
            dst_filehandler = hashing_temp_file
        try:
            while True:
                chunk = src_filehandler.read(CHUNK_SIZE)
                if not chunk:
                    break
                dst_filehandler.write(chunk)
        finally:
            dst_filehandler.close()
            # gzip does not automatically close the underlying file handler, let's do it manually
            hashing_temp_file.close()

        digest = f"sha256:{hashing_temp_file.hexdigest}"
        return hashing_temp_file.name, hashing_temp_file.total_length, digest

    def _upload_blob(self, filepath: str, size: int, digest: str) -> None:
        """Upload the blob (if necessary)."""
        # if it's already uploaded, nothing to do
        if self.registry.is_blob_already_uploaded(digest):
            emit.progress("Blob was already uploaded")
        else:
            self.registry.upload_blob(filepath, size, digest)

        # finally remove the temp filepath
        os.unlink(filepath)

    def upload_from_local(self, image_info: dict[str, Any]) -> str | None:
        """Upload the image from the local registry.

        Returns the new remote digest.
        """
        dockerd = LocalDockerdInterface()
        local_image_size = image_info["Size"]
        image_id = image_info["Id"]

        emit.progress(f"Getting the image from the local repo; size={local_image_size}")
        response = dockerd.get_streamed_image_content(image_id)

        tmp_exported = tempfile.NamedTemporaryFile(mode="wb", delete=False)
        with emit.progress_bar("Reading image...", local_image_size) as progress:
            for chunk in response.iter_content(CHUNK_SIZE):
                progress.advance(len(chunk))
                tmp_exported.file.write(chunk)
        tmp_exported.close()

        # open the image tar and inspect it to get the config and layers from the only one
        # manifest inside (as it's a list of one)
        image_tar = tarfile.open(tmp_exported.name)
        local_manifest = json.load(image_tar.extractfile("manifest.json"))
        (local_manifest,) = local_manifest
        config_name = local_manifest.get("Config")
        layer_names = local_manifest["Layers"]
        manifest = {
            "mediaType": MANIFEST_V2_MIMETYPE,
            "schemaVersion": 2,
        }

        if config_name is not None:
            fpath, size, digest = self._extract_file(image_tar, config_name)
            emit.progress(f"Uploading config blob, size={size}, digest={digest}")
            self._upload_blob(fpath, size, digest)
            manifest["config"] = {
                "digest": digest,
                "mediaType": CONFIG_MIMETYPE,
                "size": size,
            }

        manifest["layers"] = manifest_layers = []
        len_layers = len(layer_names)
        for idx, layer_name in enumerate(layer_names, 1):
            fpath, size, digest = self._extract_file(image_tar, layer_name, compress=True)
            emit.progress(f"Uploading layer blob {idx}/{len_layers}, size={size}, digest={digest}")
            self._upload_blob(fpath, size, digest)
            manifest_layers.append(
                {
                    "digest": digest,
                    "mediaType": LAYER_MIMETYPE,
                    "size": size,
                }
            )

        # remove the temp tar file
        os.unlink(tmp_exported.name)

        # upload the manifest
        manifest_data = json.dumps(manifest)
        digest = "sha256:{}".format(hashlib.sha256(manifest_data.encode("utf8")).hexdigest())
        self.registry.upload_manifest(manifest_data, digest)
        return digest



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Service class for packing."""
from __future__ import annotations

import json
import os
import pathlib
import shutil
from collections.abc import Iterable
from typing import TYPE_CHECKING, cast

import craft_application
import yaml
from craft_application import services, util
from craft_cli import emit
from craft_providers import bases

import charmcraft
from charmcraft import const, errors, models, utils
from charmcraft.models import lint
from charmcraft.models.manifest import Attribute, Manifest
from charmcraft.models.metadata import BundleMetadata, CharmMetadata
from charmcraft.models.project import (
    BasesCharm,
    Bundle,
    CharmcraftProject,
    PlatformCharm,
)

if TYPE_CHECKING:
    from charmcraft.services import CharmcraftServiceFactory
else:
    CharmcraftServiceFactory = "CharmcraftServiceFactory"


class PackageService(services.PackageService):
    """Business logic for creating packages."""

    _project: models.CharmcraftProject  # type: ignore[assignment]
    _services: CharmcraftServiceFactory

    def __init__(
        self,
        app: craft_application.AppMetadata,
        project: CharmcraftProject,
        services: CharmcraftServiceFactory,
        *,
        project_dir: pathlib.Path,
        build_plan: list[craft_application.models.BuildInfo],
    ) -> None:
        super().__init__(app, services, project=cast(craft_application.models.Project, project))
        self.project_dir = project_dir.resolve(strict=True)
        self._platform = build_plan[0].platform
        self._build_plan = build_plan

    def pack(self, prime_dir: pathlib.Path, dest: pathlib.Path) -> list[pathlib.Path]:
        """Create one or more packages as appropriate.

        :param prime_dir: Directory path to the prime directory.
        :param dest: Directory into which to write the package(s).
        :returns: A list of paths to created packages.
        """
        if self._project.type == "charm":
            packages = [self.pack_charm(prime_dir, dest)]
        elif self._project.type == "bundle":
            packages = [self.pack_bundle(prime_dir, dest)]
        else:
            raise NotImplementedError(f"Unknown package type {self._project.type}")

        self._write_package_paths(packages)
        return packages

    def _write_package_paths(self, packages: Iterable[pathlib.Path]) -> None:
        """Write the paths of packages to a hidden file in the project directory.

        This allows Charmcraft to output the packages to arbitrary directories on the host.
        """
        packages_file = self.project_dir / ".charmcraft_output_packages.txt"

        with packages_file.open("at") as file:
            file.writelines(f"{package.name}\n" for package in packages)

    def pack_bundle(self, prime_dir: pathlib.Path, dest_dir: pathlib.Path) -> pathlib.Path:
        """Pack a prime directory as a bundle."""
        name = self._project.name or "bundle"
        bundle_path = dest_dir / f"{name}.zip"
        emit.progress(f"Packing bundle {bundle_path.name}")
        utils.build_zip(bundle_path, prime_dir)
        return bundle_path

    def pack_charm(self, prime_dir: pathlib.Path, dest_dir: pathlib.Path) -> pathlib.Path:
        """Pack a prime directory as a charm for a given set of bases."""
        charm_path = self.get_charm_path(dest_dir)
        emit.progress(f"Packing charm {charm_path.name}")
        utils.build_zip(charm_path, prime_dir)

        return charm_path

    def get_charm_path(self, dest_dir: pathlib.Path) -> pathlib.Path:
        """Get a charm file name for the appropriate set of run-on bases."""
        if self._platform:
            return dest_dir / f"{self._project.name}_{self._platform}.charm"
        build_plan = models.CharmcraftBuildPlanner.model_validate(
            self._project.marshal()
        ).get_build_plan()
        platform = utils.get_os_platform()
        build_on_base = bases.BaseName(name=platform.system, version=platform.release)
        host_arch = util.get_host_architecture()
        for build_info in build_plan:
            print(build_info)
            if build_info.build_on != host_arch:
                continue
            if build_info.base == build_on_base:
                return dest_dir / f"{self._project.name}_{build_info.platform}.charm"

        raise errors.CraftError(
            "Current machine is not a valid build platform for this charm.",
        )

    @property
    def metadata(self) -> BundleMetadata | CharmMetadata:
        """Metadata model for this project."""
        if isinstance(self._project, BasesCharm | PlatformCharm):
            return CharmMetadata.from_charm(self._project)
        if isinstance(self._project, Bundle):
            return BundleMetadata.from_bundle(self._project)
        raise NotImplementedError(f"Unknown project type {self._project.type!r}")

    def _write_file_or_object(
        self, model: dict | None, filename: str, dest_dir: pathlib.Path
    ) -> None:
        """Write a yaml file to the destination directory if the given object is not None.

        This function prefers copying the file over, but will generate YAML from the given
        model otherwise.

        :param model: The dictionary to write (or None)
        :param filename: The name of the file to copy or write.
        :param dest_dir: The path of the destination directory.
        """
        if not model:
            return
        dest_dir.mkdir(parents=True, exist_ok=True)
        source_path = self.project_dir / filename
        dest_path = dest_dir / filename
        if source_path.is_file():
            shutil.copyfile(source_path, dest_path)
            return
        with dest_path.open("wt+") as dest_file:
            yaml.safe_dump(model, dest_file)

    def get_manifest(self, lint_results: Iterable[lint.CheckResult]) -> Manifest:
        """Get the manifest for this charm."""
        attributes = [
            Attribute(name=result.name, result=result.result)
            for result in lint_results
            if result.check_type == lint.CheckType.ATTRIBUTE
        ]

        if image_info := os.getenv(const.IMAGE_INFO_ENV_VAR):
            try:
                image_info = json.loads(image_info)
            except json.decoder.JSONDecodeError as exc:
                msg = f"Failed to parse the content of {const.IMAGE_INFO_ENV_VAR} environment variable"
                raise errors.CraftError(msg) from exc

        bases = self.get_manifest_bases()

        return Manifest(
            charmcraft_version=charmcraft.__version__,
            charmcraft_started_at=self._project.started_at.isoformat(),
            analysis={"attributes": attributes},
            image_info=image_info,
            bases=bases,
        )

    def get_manifest_bases(self) -> list[models.Base]:
        """Get the bases used for a charm manifest from the project."""
        if isinstance(self._project, BasesCharm):
            run_on_bases = []
            for project_base in self._project.bases:
                for build_base in project_base.build_on:
                    if build_base.name != self._build_plan[0].base.name:
                        continue
                    if build_base.channel != self._build_plan[0].base.version:
                        continue
                    if self._build_plan[0].build_on not in build_base.architectures:
                        continue
                    run_on_bases.extend(project_base.run_on)
            if not run_on_bases:
                raise RuntimeError("Could not determine run-on bases.")
            return run_on_bases
        if isinstance(self._project, PlatformCharm):
            if not self._platform:
                architectures = [util.get_host_architecture()]
            elif self._platform in (*const.SUPPORTED_ARCHITECTURES, "all"):
                architectures = [self._platform]
            elif platform := self._project.platforms.get(self._platform):
                if platform.build_for:
                    architectures = [str(arch) for arch in platform.build_for]
                else:
                    raise ValueError(f"Platform {self._platform} contains unknown build-for.")
            else:
                architectures = [util.get_host_architecture()]
            return [models.Base.from_str_and_arch(self._project.base, architectures)]
        raise TypeError(f"Unknown charm type {self._project.__class__}, cannot get bases.")

    def write_metadata(self, path: pathlib.Path) -> None:
        """Write additional charm metadata.

        :param path: The path to the prime directory.
        """
        path.mkdir(parents=True, exist_ok=True)
        if isinstance(self._project, BasesCharm | PlatformCharm):
            if self._project.analysis:
                ignore_checkers = {
                    *self._project.analysis.ignore.linters,
                    *self._project.analysis.ignore.attributes,
                }
            else:
                ignore_checkers = set()
            lint_results = self._services.analysis.lint_directory(
                self._services.lifecycle.prime_dir, ignore=ignore_checkers
            )
            manifest = self.get_manifest(lint_results)
            # Converting the manifest to a dictionary here is fairly fragile.
            # We need to include unset/default values in order to ensure that the
            # architecture is included on each base in the manifest, even when the
            # architectures are inferred. However, we also need to exclude Nones so that
            # image-info isn't included in manifest.yaml if it doesn't exist.
            # Tread carefully when changing this next line. Treat it like an antique
            # crystal wine glass.
            (path / "manifest.yaml").write_text(
                utils.dump_yaml(
                    manifest.model_dump(
                        mode="json", by_alias=True, exclude_unset=False, exclude_none=True
                    )
                )
            )

        project_dict = self._project.marshal()

        # If there is a reactive part, defer to it for the existence of metadata.yaml.
        plugins = {part.get("plugin") or name for name, part in self._project.parts.items()}
        is_reactive = "reactive" in plugins
        stage_dir = self._services.lifecycle.project_info.dirs.stage_dir
        if is_reactive and (stage_dir / const.METADATA_FILENAME).exists():
            emit.debug(
                f"{const.METADATA_FILENAME!r} generated by charm. Not using original project metadata."
            )
        else:
            self._write_file_or_object(self.metadata.marshal(), const.METADATA_FILENAME, path)
        if is_reactive and (stage_dir / const.JUJU_ACTIONS_FILENAME).exists():
            emit.debug(f"{const.JUJU_ACTIONS_FILENAME!r} generated by charm. Skipping generation.")
        elif actions := cast(dict | None, project_dict.get("actions")):
            self._write_file_or_object(actions, "actions.yaml", path)
        if config := cast(dict | None, project_dict.get("config")):
            self._write_file_or_object(config, "config.yaml", path)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for project-related models."""

import itertools
import json
import pathlib
import textwrap
from textwrap import dedent
from typing import Any

import hypothesis
import pydantic
import pyfakefs.fake_filesystem
import pytest
import pytest_check
from craft_application import models, util
from craft_application.errors import CraftValidationError
from craft_application.util import safe_yaml_load
from craft_cli import CraftError
from craft_providers import bases
from hypothesis import strategies

from charmcraft import const, utils
from charmcraft.models import project
from charmcraft.models.charmcraft import Base, BasesConfiguration

SIMPLE_BASE = Base(name="simple", channel="0.0")
BASE_WITH_ONE_ARCH = Base(name="arch", channel="1.0", architectures=["amd64"])
BASE_WITH_MULTIARCH = Base(name="multiarch", channel="2.0", architectures=["arm64", "riscv64"])
SIMPLE_BASENAME = bases.BaseName("simple", "0.0")
ONE_ARCH_BASENAME = bases.BaseName("arch", "1.0")
MULTIARCH_BASENAME = bases.BaseName("multiarch", "2.0")
UBUNTU_JAMMY = Base(name="ubuntu", channel="22.04", architectures=["amd64"])
SIMPLE_BASES = (
    UBUNTU_JAMMY,
    Base(name="centos", channel="7", architectures=["amd64"]),
    Base(name="almalinux", channel="9", architectures=["amd64"]),
)
COMPLEX_BASES = (
    Base(
        name="ubuntu",
        channel="devel",
        architectures=["amd64", "arm64", "s390x", "riscv64"],
    ),
    Base(
        name="almalinux",
        channel="9",
        architectures=["amd64", "arm64", "s390x", "ppc64el"],
    ),
)
SIMPLE_BASE_CONFIG_DICT = {"name": "ubuntu", "channel": "22.04"}
FULL_BASE_CONFIG_DICT = {
    "build-on": [{"channel": "22.04", "name": "ubuntu"}],
    "run-on": [{"channel": "22.04", "name": "ubuntu"}],
}
BASIC_CHARM_PARTS = {"charm": {"plugin": "charm", "source": "."}}
BASIC_CHARM_PARTS_EXPANDED = {
    "charm": {
        "plugin": "charm",
        "source": ".",
        "charm-binary-python-packages": [],
        "charm-entrypoint": "src/charm.py",
        "charm-python-packages": [],
        "charm-requirements": [],
        "charm-strict-dependencies": False,
    }
}

MINIMAL_CHARMCRAFT_YAML = f"""\
type: charm
bases:
  - build-on:
      - name: ubuntu
        channel: "22.04"
        architectures: [{util.get_host_architecture()}]
    run-on:
      - name: ubuntu
        channel: "22.04"
        architectures: [arm64]
"""
SIMPLE_METADATA_YAML = "{name: charmy-mccharmface, summary: Charmy!, description: Very charming!}"
SIMPLE_CHARMCRAFT_YAML = f"""\
type: charm
name: charmy-mccharmface
summary: Charmy!
description: Very charming!
bases:
  - build-on:
      - name: ubuntu
        channel: "22.04"
        architectures: [{util.get_host_architecture()}]
    run-on:
      - name: ubuntu
        channel: "22.04"
        architectures: [arm64]
"""
SIMPLE_CONFIG_YAML = "options: {admin: {default: root, description: Admin user, type: string}}"
SIMPLE_CONFIG_DICT = {
    "options": {"admin": {"type": "string", "default": "root", "description": "Admin user"}}
}
SIMPLE_ACTIONS_YAML = "snooze: {description: Take a little nap.}"
SIMPLE_ACTIONS_DICT = {"snooze": {"description": "Take a little nap."}}


# region CharmPlatform tests
@pytest.mark.parametrize(
    ("bases", "expected"),
    [
        (
            [Base(name="xname", channel="xchannel", architectures=["xarch1"])],
            "xname-xchannel-xarch1",
        ),
        (
            [
                Base(
                    name="xname",
                    channel="xchannel",
                    architectures=["xarch1", "xarch2", "xarch3"],
                )
            ],
            "xname-xchannel-xarch1-xarch2-xarch3",
        ),
        (
            [
                Base(name="x1name", channel="x1channel", architectures=["x1arch"]),
                Base(
                    name="x2name",
                    channel="x2channel",
                    architectures=["x2arch1", "x2arch2"],
                ),
            ],
            "x1name-x1channel-x1arch_x2name-x2channel-x2arch1-x2arch2",
        ),
    ],
)
def test_platform_from_bases_backwards_compatible(bases: list[Base], expected: str):
    """Replicates the format_charm_file_name tests in test_package.py.

    This ensures that charm names remain consistent as we move to platforms.
    """
    assert project.get_charm_file_platform_str(bases) == expected


@pytest.mark.parametrize("base", [*SIMPLE_BASES, *COMPLEX_BASES])
def test_platform_from_single_base(base):
    """Tests generating a platform name from a real (single) base."""
    expected_architectures = "-".join(base.architectures)
    expected = f"{base.name}-{base.channel}-{expected_architectures}"

    actual = project.get_charm_file_platform_str([base])

    assert actual == expected


@pytest.mark.parametrize(
    ("bases", "expected"),
    [
        (SIMPLE_BASES, "ubuntu-22.04-amd64_centos-7-amd64_almalinux-9-amd64"),
        (
            COMPLEX_BASES,
            "ubuntu-devel-amd64-arm64-s390x-riscv64_almalinux-9-amd64-arm64-s390x-ppc64el",
        ),
    ],
)
def test_platform_from_multiple_bases(bases, expected):
    assert project.get_charm_file_platform_str(bases) == expected


# endregion
# region Platform tests
VALID_PLATFORM_ARCHITECTURES = [
    *(
        list(x) for x in itertools.combinations(const.CharmArch, 1)
    ),  # A single architecture in a list
    *(list(x) for x in itertools.combinations(const.CharmArch, 2)),  # Two architectures in a list
]


# endregion
# region CharmBuildInfo tests
@pytest.mark.parametrize("build_on_base", [SIMPLE_BASE, BASE_WITH_ONE_ARCH, BASE_WITH_MULTIARCH])
@pytest.mark.parametrize("build_on_arch", ["amd64", "arm64", "riscv64", "s390x"])
@pytest.mark.parametrize("run_on", [SIMPLE_BASE, BASE_WITH_ONE_ARCH])
def test_build_info_from_build_on_run_on_basic(
    build_on_base: Base,
    build_on_arch: str,
    run_on: Base,
):
    info = project.CharmBuildInfo.from_build_on_run_on(
        build_on_base, build_on_arch, [run_on], bases_index=10, build_on_index=256
    )

    pytest_check.equal(info.build_on, build_on_arch)
    pytest_check.equal(info.build_for_bases, [run_on])
    pytest_check.equal(info.build_for, run_on.architectures[0])
    pytest_check.equal(info.base.name, build_on_base.name)
    pytest_check.equal(info.base.version, build_on_base.channel)


@pytest.mark.parametrize(
    ("lib_name", "expected_lib_name"),
    [
        ("charm.lib", "charm.lib"),
        ("charm_with_underscores.lib", "charm-with-underscores.lib"),
        ("charm-with-hyphens.lib", "charm-with-hyphens.lib"),
        ("charm.lib_with_hyphens", "charm.lib_with_hyphens"),
        ("charm0.number_0_lib", "charm0.number_0_lib"),
    ],
)
@pytest.mark.parametrize("lib_version", ["0", "1", "2.0", "2.1", "3.14"])
def test_create_valid_charm_lib(lib_name: str, expected_lib_name: str, lib_version: str):
    lib = project.CharmLib.unmarshal({"lib": lib_name, "version": lib_version})
    assert lib.lib == expected_lib_name


@pytest.mark.parametrize(
    ("name", "error_match"),
    [
        (
            "boop",
            r"Library name invalid. Expected '\[charm_name\].\[lib_name\]', got 'boop'",
        ),
        (
            "Invalid charm name.valid_lib",
            "Invalid charm name for lib 'Invalid charm name.valid_lib'. Value 'Invalid charm name' is invalid",
        ),
        (
            "my_charm.invalid-library-name",
            "Library name 'invalid-library-name' is invalid. Library names must be valid Python module names.",
        ),
    ],
)
def test_invalid_charm_lib_name(name: str, error_match: str):
    with pytest.raises(pydantic.ValidationError, match=error_match):
        project.CharmLib.unmarshal({"lib": name, "version": "0"})


@hypothesis.given(
    strategies.one_of(
        strategies.floats(
            min_value=0.001,
            max_value=2**32,
            allow_nan=False,
            allow_infinity=False,
            allow_subnormal=False,
        ),
        strategies.integers(min_value=0, max_value=2**32),
    )
)
def test_valid_library_version(version: float):
    project.CharmLib.unmarshal({"lib": "charm_name.lib_name", "version": str(version)})


@pytest.mark.parametrize("version", [".1", "NaN", ""])
def test_invalid_api_version(version: str):
    with pytest.raises(
        pydantic.ValidationError,
        match="API version not valid. Expected an integer, got '",
    ):
        project.CharmLib(lib="charm_name.lib_name", version=version)


@pytest.mark.parametrize("version", ["1.", "1.number"])
def test_invalid_patch_version(version: str):
    with pytest.raises(
        pydantic.ValidationError,
        match="Patch version not valid. Expected an integer, got '",
    ):
        project.CharmLib(lib="charm_name.lib_name", version=version)


@pytest.mark.parametrize(
    ("run_on", "expected"),
    [
        pytest.param(
            [BASE_WITH_MULTIARCH],
            "arm64-riscv64",
            id="one-base",
        ),
        pytest.param(
            [
                BASE_WITH_ONE_ARCH,
                Base(name="ubuntu", channel="24.04", architectures=["riscv64"]),
            ],
            "amd64-riscv64",
            id="multiarch-across-bases",
        ),
    ],
)
def test_build_info_from_build_on_run_on_multi_arch(run_on, expected):
    info = project.CharmBuildInfo.from_build_on_run_on(
        SIMPLE_BASE, "amd64", run_on, bases_index=10, build_on_index=256
    )

    assert info.build_for == expected


@pytest.mark.parametrize(
    ("given", "expected"),
    [
        pytest.param([], [], id="empty"),
        pytest.param(
            [
                BasesConfiguration(
                    **{"build-on": [BASE_WITH_ONE_ARCH], "run-on": [BASE_WITH_ONE_ARCH]}
                )
            ],
            [
                project.CharmBuildInfo(
                    platform="arch-1.0-amd64",  # mypy: ignore[assignment]
                    build_on="amd64",
                    build_for="amd64",
                    base=ONE_ARCH_BASENAME,
                    build_for_bases=[BASE_WITH_ONE_ARCH],
                    bases_index=0,
                    build_on_index=0,
                )
            ],
        ),
    ],
)
def test_build_info_generator(given, expected):
    assert list(project.CharmBuildInfo.gen_from_bases_configurations(*given)) == expected


# endregion
# region Build planners
@pytest.mark.parametrize(
    ("data", "expected"),
    [
        pytest.param(
            {"type": "bundle"},
            [
                project.models.BuildInfo(
                    platform=util.get_host_architecture(),
                    build_on=util.get_host_architecture(),
                    build_for=util.get_host_architecture(),
                    base=bases.BaseName(
                        name=utils.get_os_platform().system,
                        version=utils.get_os_platform().release,
                    ),
                ),
            ],
            id="bundle",
        ),
        pytest.param(
            {"base": "ubuntu@22.04", "platforms": {"amd64": None}},
            [
                project.models.BuildInfo(
                    platform="amd64",
                    build_on="amd64",
                    build_for="amd64",
                    base=bases.BaseName("ubuntu", "22.04"),
                )
            ],
            id="simple-platforms",
        ),
        pytest.param(
            {
                "base": "ubuntu@22.04",
                "platforms": {
                    "fancy": {
                        "build-on": ["amd64", "arm64", "riscv64"],
                        "build-for": ["s390x"],
                    },
                    "crossy": {"build-on": ["s390x"], "build-for": ["ppc64el"]},
                    "amd64": None,
                    "arm64": None,
                    "riscv64": None,
                },
            },
            [
                project.models.BuildInfo(
                    platform="fancy",
                    build_on="amd64",
                    build_for="s390x",
                    base=bases.BaseName("ubuntu", "22.04"),
                ),
                project.models.BuildInfo(
                    platform="fancy",
                    build_on="arm64",
                    build_for="s390x",
                    base=bases.BaseName("ubuntu", "22.04"),
                ),
                project.models.BuildInfo(
                    platform="fancy",
                    build_on="riscv64",
                    build_for="s390x",
                    base=bases.BaseName("ubuntu", "22.04"),
                ),
                project.models.BuildInfo(
                    platform="crossy",
                    build_on="s390x",
                    build_for="ppc64el",
                    base=bases.BaseName("ubuntu", "22.04"),
                ),
                project.models.BuildInfo(
                    platform="amd64",
                    build_on="amd64",
                    build_for="amd64",
                    base=bases.BaseName("ubuntu", "22.04"),
                ),
                project.models.BuildInfo(
                    platform="arm64",
                    build_on="arm64",
                    build_for="arm64",
                    base=bases.BaseName("ubuntu", "22.04"),
                ),
                project.models.BuildInfo(
                    platform="riscv64",
                    build_on="riscv64",
                    build_for="riscv64",
                    base=bases.BaseName("ubuntu", "22.04"),
                ),
            ],
            id="complex-platforms",
        ),
        pytest.param(
            {
                "base": "ubuntu@24.04",
                "build-base": "ubuntu@devel",
                "platforms": {"amd64": None},
            },
            [
                project.models.BuildInfo(
                    platform="amd64",
                    build_on="amd64",
                    build_for="amd64",
                    base=bases.BaseName("ubuntu", "devel"),
                )
            ],
            id="platforms-with-build-base",
        ),
        pytest.param(
            {"bases": [{"name": "ubuntu", "channel": "22.04"}]},
            [
                project.CharmBuildInfo(
                    platform=f"ubuntu-22.04-{util.get_host_architecture()}",
                    build_on=util.get_host_architecture(),
                    build_for=util.get_host_architecture(),
                    build_for_bases=[project.charmcraft.Base(name="ubuntu", channel="22.04")],
                    build_on_index=0,
                    base=bases.BaseName("ubuntu", "22.04"),
                    bases_index=0,
                ),
            ],
            id="basic-bases",
        ),
        pytest.param(
            {"bases": [{"build-on": [BASE_WITH_ONE_ARCH], "run-on": [BASE_WITH_ONE_ARCH]}]},
            [
                project.CharmBuildInfo(
                    platform="arch-1.0-amd64",
                    build_on="amd64",
                    build_for="amd64",
                    build_for_bases=[BASE_WITH_ONE_ARCH],
                    build_on_index=0,
                    base=bases.BaseName("arch", "1.0"),
                    bases_index=0,
                ),
            ],
            id="arch-base",
        ),
    ],
)
def test_build_planner_correct(data, expected):
    planner = project.CharmcraftBuildPlanner.model_validate(data)

    assert planner.get_build_plan() == expected


@pytest.mark.parametrize("base", ["ubuntu@20.04", "ubuntu@22.04", "ubuntu@24.04"])
@pytest.mark.parametrize(
    ("build_base", "build_plan_basename"),
    [
        ("ubuntu@20.04", bases.BaseName("ubuntu", "20.04")),
        ("ubuntu@22.04", bases.BaseName("ubuntu", "22.04")),
        ("ubuntu@24.04", bases.BaseName("ubuntu", "24.04")),
        ("almalinux@9", bases.BaseName("almalinux", "9")),
    ],
)
@pytest.mark.parametrize(
    "platforms",
    [
        {"amd64": None},
        {
            "amd64": None,
            "fancy": {"build-on": ["amd64", "riscv64"], "build-for": ["riscv64"]},
        },
    ],
)
def test_build_planner_platforms_combinations(base, build_base, build_plan_basename, platforms):
    """Test that we're able to create a valid platform for each of these combinations."""
    planner = project.CharmcraftBuildPlanner(
        base=base,
        build_base=build_base,
        platforms=platforms,
    )
    plan = planner.get_build_plan()

    for build_info in plan:
        pytest_check.equal(build_info.base, build_plan_basename)
        pytest_check.is_in(build_info.platform, platforms.keys())


@pytest.mark.parametrize("architecture", sorted(const.SUPPORTED_ARCHITECTURES))
@pytest.mark.parametrize("system", ["ubuntu", "linux", "macos", "windows", "plan9"])
@pytest.mark.parametrize("release", ["22.04", "2.6.32", "10.5", "vista", "from bell labs"])
def test_get_bundle_plan(mocker, architecture, release, system):
    mocker.patch("craft_application.util.get_host_architecture", return_value=architecture)
    mocker.patch(
        "charmcraft.utils.get_os_platform",
        return_value=utils.OSPlatform(machine=architecture, system=system, release=release),
    )
    planner = project.CharmcraftBuildPlanner(type="bundle")

    assert planner.get_build_plan() == [
        models.BuildInfo(
            platform=architecture,
            build_on=architecture,
            build_for=architecture,
            base=bases.BaseName(system, release),
        )
    ]


# endregion
# region CharmcraftProject tests
@pytest.mark.parametrize(
    ("data", "type_class"),
    [
        (
            {
                "type": "charm",
                "name": "basic",
                "summary": "A basic charm",
                "description": "A thing",
                "bases": [SIMPLE_BASE_CONFIG_DICT],
            },
            project.Charm,
        ),
        (
            {"type": "bundle"},
            project.Bundle,
        ),
    ],
)
def test_unmarshal_success(data, type_class):
    assert isinstance(project.CharmcraftProject.unmarshal(data), type_class)


@pytest.mark.parametrize("type_", [None, "", "invalid", "Dvorak"])
def test_unmarshal_invalid_type(type_):
    with pytest.raises(ValueError, match="^field type cannot be "):
        project.CharmcraftProject.unmarshal({"type": type_})


@pytest.mark.parametrize(
    (
        "charmcraft_yaml",
        "metadata_yaml",
        "config_yaml",
        "actions_yaml",
        "expected_diff",
    ),
    [
        (
            SIMPLE_CHARMCRAFT_YAML,
            None,
            None,
            None,
            {"parts": BASIC_CHARM_PARTS},
        ),
        (
            MINIMAL_CHARMCRAFT_YAML,
            SIMPLE_METADATA_YAML,
            None,
            None,
            {"parts": BASIC_CHARM_PARTS},
        ),
        (
            SIMPLE_CHARMCRAFT_YAML,
            None,
            SIMPLE_CONFIG_YAML,
            None,
            {"config": SIMPLE_CONFIG_DICT, "parts": BASIC_CHARM_PARTS},
        ),
        (
            SIMPLE_CHARMCRAFT_YAML,
            None,
            None,
            SIMPLE_ACTIONS_YAML,
            {"actions": SIMPLE_ACTIONS_DICT, "parts": BASIC_CHARM_PARTS},
        ),
        (
            MINIMAL_CHARMCRAFT_YAML,
            SIMPLE_METADATA_YAML,
            SIMPLE_CONFIG_YAML,
            SIMPLE_ACTIONS_YAML,
            {
                "actions": SIMPLE_ACTIONS_DICT,
                "config": SIMPLE_CONFIG_DICT,
                "parts": BASIC_CHARM_PARTS,
            },
        ),
        pytest.param(
            SIMPLE_CHARMCRAFT_YAML
            + textwrap.dedent(
                """\
                parts:
                  charm: {}
                  reactive: {}
                  bundle: {}
                """
            ),
            None,
            None,
            None,
            {
                "parts": {
                    "charm": {
                        "plugin": "charm",
                        "source": ".",
                    },
                    "reactive": {
                        "plugin": "reactive",
                    },
                    "bundle": {
                        "plugin": "bundle",
                        "source": ".",
                    },
                }
            },
            id="implicit-parts-plugins",
        ),
    ],
)
def test_from_yaml_file_success(
    fs: pyfakefs.fake_filesystem.FakeFilesystem,
    simple_charm,
    charmcraft_yaml: str,
    metadata_yaml: str | None,
    config_yaml: str | None,
    actions_yaml: str | None,
    expected_diff: dict[str, Any],
):
    expected_dict = simple_charm.marshal().copy()
    expected_dict.update(expected_diff)

    fs.create_file("/charmcraft.yaml", contents=charmcraft_yaml)
    if metadata_yaml:
        fs.create_file("/metadata.yaml", contents=metadata_yaml)
    if config_yaml:
        fs.create_file("/config.yaml", contents=config_yaml)
    if actions_yaml:
        fs.create_file("/actions.yaml", contents=actions_yaml)

    actual = project.CharmcraftProject.from_yaml_file(pathlib.Path("/charmcraft.yaml"))

    assert actual.marshal() == expected_dict


@pytest.mark.parametrize(
    (
        "charmcraft_yaml",
        "metadata_yaml",
        "config_yaml",
        "actions_yaml",
        "exc_class",
        "match",
        "details",
    ),
    [
        pytest.param(
            None,
            None,
            None,
            None,
            CraftError,
            r"^Could not find charmcraft\.yaml at '.charmcraft\.yaml'$",
            None,
            id="FileNotFound",
        ),
        pytest.param(
            f"{SIMPLE_CHARMCRAFT_YAML}\nconfig: ",
            None,
            SIMPLE_CONFIG_YAML,
            None,
            CraftValidationError,
            r"^Cannot specify 'config' section",
            None,
            id="duplicate-config",
        ),
        pytest.param(
            f"{SIMPLE_CHARMCRAFT_YAML}\nactions:",
            None,
            None,
            SIMPLE_ACTIONS_YAML,
            CraftValidationError,
            r"^Cannot specify 'actions' section",
            None,
            id="duplcate-actions",
        ),
    ],
)
def test_from_yaml_file_exception(
    fs: pyfakefs.fake_filesystem.FakeFilesystem,
    charmcraft_yaml: str | None,
    metadata_yaml: str | None,
    config_yaml: str | None,
    actions_yaml: str | None,
    exc_class: type[CraftError],
    match: str,
    details: str,
):
    if charmcraft_yaml:
        fs.create_file("/charmcraft.yaml", contents=charmcraft_yaml)
    if metadata_yaml:
        fs.create_file("/metadata.yaml", contents=metadata_yaml)
    if config_yaml:
        fs.create_file("/config.yaml", contents=config_yaml)
    if actions_yaml:
        fs.create_file("/actions.yaml", contents=actions_yaml)

    with pytest.raises(exc_class, match=match) as exc:
        project.CharmcraftProject.from_yaml_file(pathlib.Path("/charmcraft.yaml"))

    assert exc.value.details == details


# endregion
# region Charm tests
@pytest.mark.parametrize(
    ("values", "expected_changes"),
    [
        pytest.param(
            {"bases": [SIMPLE_BASE_CONFIG_DICT]},
            {"bases": [FULL_BASE_CONFIG_DICT]},
            id="simple-base",
        ),
    ],
)
def test_instantiate_bases_charm_success(values: dict[str, Any], expected_changes: dict[str, Any]):
    """Various successful instantiations of a charm project."""
    values.update(
        {
            "type": "charm",
            "name": "test-charm",
            "summary": "A test charm",
            "description": "A test charm model that can be successfully instantiated.",
        }
    )
    expected = values.copy()
    expected.update(expected_changes)

    actual = project.BasesCharm(**values)

    assert actual.marshal() == expected


@pytest.mark.parametrize(
    ("values", "error_cls", "error_match"),
    [
        pytest.param(
            {
                "type": "charm",
                "name": "test-charm",
                "summary": "A test charm",
                "description": "This charm has no bases and is thus invalid.",
            },
            pydantic.ValidationError,
            r"bases\s+Field required",
            id="no-bases",
        ),
        pytest.param(
            {
                "type": "charm",
                "name": "test-charm",
                "summary": "A test charm",
                "description": "Empty bases, also invalid.",
                "bases": [],
            },
            pydantic.ValidationError,
            r"bases\s+List should have at least 1 item",
            id="empty-bases",
        ),
    ],
)
def test_instantiate_bases_charm_error(
    values: dict[str, Any], error_cls: type[Exception], error_match: str
):
    with pytest.raises(error_cls, match=error_match):
        project.BasesCharm(**values)


@pytest.mark.parametrize("base", ["ubuntu@18.04", "ubuntu@22.04"])
def test_devel_bases(monkeypatch, base):
    monkeypatch.setattr(const, "DEVEL_BASE_STRINGS", [base])

    with pytest.raises(
        pydantic.ValidationError,
        match=r"requires a build-base \(recommended: 'build-base: ubuntu@devel'\)",
    ):
        project.PlatformCharm(
            type="charm",
            name="test-charm",
            summary="",
            description="",
            base=base,
            platforms={"amd64": None},
            parts={"charm": {"plugin": "charm"}},
        )


@pytest.mark.parametrize(
    "filename",
    [f.name for f in (pathlib.Path(__file__).parent / "valid_charms_yaml").iterdir()],
)
def test_read_charm_from_yaml_file_self_contained_success(tmp_path, filename: str):
    file_path = pathlib.Path(__file__).parent / "valid_charms_yaml" / filename
    with file_path.open() as f:
        expected_dict = safe_yaml_load(f)

    charm = project.CharmcraftProject.from_yaml_file(file_path)

    # The JSON round-trip here is to get rid of any Pydantic constraint objects like AnyHttpUrl
    assert json.loads(json.dumps(charm.marshal())) == expected_dict


@pytest.mark.parametrize(
    ("filename", "errors"),
    [
        (
            "basic.yaml",
            dedent(
                """\
                Bad basic.yaml content:
                - field 'name' required in top-level configuration
                - field 'summary' required in top-level configuration
                - field 'description' required in top-level configuration
                - field 'bases' required in top-level configuration"""
            ),
        ),
        (
            "invalid-type.yaml",
            dedent(
                """\
                Bad invalid-type.yaml content:
                - field 'name' required in top-level configuration
                - field 'summary' required in top-level configuration
                - field 'description' required in top-level configuration
                - input should be 'charm' (in field 'type')
                - field 'bases' required in top-level configuration"""
            ),
        ),
        (
            "invalid-base.yaml",
            dedent(
                """\
                Bad invalid-base.yaml content:
                - base requires 'platforms' definition: {'name': 'ubuntu', 'channel': '24.04'} (in field 'bases[0]')
                - base requires 'platforms' definition: {'name': 'ubuntu', 'channel': 'devel'} (in field 'bases[1]')"""
            ),
        ),
    ],
)
def test_read_charm_from_yaml_file_error(filename, errors):
    file_path = pathlib.Path(__file__).parent / "invalid_charms_yaml" / filename

    with pytest.raises(CraftValidationError) as exc:
        _ = project.BasesCharm.from_yaml_file(file_path)

    assert exc.value.args[0] == errors


@pytest.mark.parametrize(
    ("base", "expected"),
    [
        ({"name": "ubuntu", "channel": "18.04"}, True),
        ({"name": "ubuntu", "channel": "20.04"}, True),
        ({"name": "ubuntu", "channel": "22.04"}, True),
        ({"name": "ubuntu", "channel": "23.04"}, True),
        ({"name": "ubuntu", "channel": "23.10"}, True),
        ({"name": "ubuntu", "channel": "24.04"}, False),
        ({"name": "ubuntu", "channel": "24.10"}, False),
        ({"name": "ubuntu", "channel": "25.04"}, False),
        ({"name": "centos", "channel": "7"}, True),
        ({"name": "almalinux", "channel": "9"}, True),
    ],
)
def test_check_legacy_bases(base, expected):
    assert project._check_base_is_legacy(base) == expected


# endregion



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""General fixtures for integration tests."""
from unittest import mock

import craft_store
import pytest

from charmcraft import application, services
from charmcraft.application import commands


@pytest.fixture
def service_factory():
    factory = services.CharmcraftServiceFactory(app=application.APP_METADATA)
    factory.store.client = mock.Mock(spec_set=craft_store.StoreClient)
    return factory


@pytest.fixture
def app(monkeypatch, new_path, service_factory):
    monkeypatch.setenv("CRAFT_DEBUG", "1")
    app = application.Charmcraft(app=application.APP_METADATA, services=service_factory)
    app._configure_services(None)
    commands.fill_command_groups(app)

    return app



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Infrastructure for the 'test' command."""
import argparse
import os
import subprocess

from craft_cli import CraftError, emit

from charmcraft import env
from charmcraft.application.commands import base

_overview = """
Run charm tests in different back-ends.

This command will run charm test suites using the spread tool. For further
information, see the spread documentation: https://github.com/snapcore/spread
"""


class TestCommand(base.CharmcraftCommand):
    """Initialize a directory to be a charm project."""

    name = "test"
    help_msg = "Execute charm test suites"
    overview = _overview
    common = True

    def fill_parser(self, parser: argparse.ArgumentParser) -> None:
        """Specify the parameters for this command."""
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--shell",
            action="store_true",
            help="Shell into the environment in lieu of the step to run.",
        )
        group.add_argument(
            "--shell-after",
            action="store_true",
            help="Shell into the environment after the step has run.",
        )

        parser.add_argument(
            "--debug",
            action="store_true",
            help="Shell into the environment if the build fails.",
        )

        parser.add_argument(
            "--list",
            action="store_true",
            help="Just show the list of jobs that would run.",
        )

        parser.add_argument(
            "spread_tasks",
            metavar="tasks",
            nargs=argparse.REMAINDER,
            help=(
                "Spread tasks to run, in backend:system:suite/task:variant "
                "format. All fields are optional."
            ),
        )

    def run(self, parsed_args: argparse.Namespace):
        """Execute command's actual functionality."""
        if env.is_charmcraft_running_from_snap():
            cmd = [f"{os.environ['SNAP']}/bin/spread"]
        else:
            cmd = ["spread"]

        for arg in ("shell", "shell-after", "debug", "list"):
            if vars(parsed_args).get(arg):
                cmd.append("-" + arg)

        # Choose the github-ci backend if running on GitHub, otherwise default
        # to multipass.
        spread_tasks = parsed_args.spread_tasks
        if len(spread_tasks) == 0:
            if os.environ.get("GITHUB_RUN_ID"):
                spread_tasks = ["github-ci"]
            else:
                spread_tasks = ["multipass"]

        try:
            with emit.pause():
                subprocess.run([*cmd, *spread_tasks], check=True)
        except subprocess.CalledProcessError as err:
            raise CraftError(f"test error: {err}")



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft Juju Actions pydantic model."""

import keyword
import re

import pydantic
from craft_application.models import CraftBaseModel


class JujuActions(CraftBaseModel):
    """Juju actions for charms.

    See also: https://juju.is/docs/sdk/actions
    """

    _action_name_regex = re.compile(r"^[a-zA-Z_][a-zA-Z0-9-_]*$")
    actions: dict[str, dict] | None

    @pydantic.field_validator("actions", mode="after")
    def validate_actions(cls, actions):
        """Verify actions names and descriptions."""
        if not isinstance(actions, dict):
            raise TypeError("actions.yaml is not a valid actions configuration")
        for action in actions:
            if keyword.iskeyword(action):
                raise ValueError(
                    f"'{action}' is a reserved keyword and cannot be used as an action name"
                )
            if cls._action_name_regex.match(action) is None:
                raise ValueError(f"'{action}' is not a valid action name")

        return actions

    @pydantic.field_validator("actions", mode="after")
    def _validate_actions(cls, action):
        """Verify actions names and descriptions."""
        if not isinstance(action, dict):
            raise TypeError(f"'{action}' is not a dictionary")

        if "description" not in action:
            raise ValueError(f"'{action}' is missing a description")

        return action



# Copyright 2023-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Charm plugin for craft-parts."""
import os
import pathlib
import re
import shlex
import sys
from contextlib import suppress
from typing import Literal, cast

import overrides
import pydantic
from craft_parts import Step, callbacks, plugins
from craft_parts.errors import OsReleaseIdError, OsReleaseVersionIdError
from craft_parts.packages import platform
from craft_parts.utils import os_utils
from typing_extensions import Self

from charmcraft import charm_builder, env, instrum

PACKAGE_NAME_REGEX = re.compile(r"[A-Za-z0-9_.-]+")


class CharmPluginProperties(plugins.PluginProperties, frozen=True):
    """Properties used in charm building."""

    plugin: Literal["charm"] = "charm"
    source: str = "."
    charm_entrypoint: str = "src/charm.py"
    charm_binary_python_packages: list[str] = []
    charm_python_packages: list[str] = []
    charm_requirements: list[str] = []
    charm_strict_dependencies: bool = False
    """Whether to select strict dependencies only.

    If true, ``charm-strict-dependencies`` will enforce that all dependencies, direct or indirect,
    be specified within a requirements file. This includes any ``PYDEPS`` specified from a charm
    library. It also changes the behaviour of ``charm-binary-python-packages`` to be a list of
    packages to pass to ``pip`` that are allowed to use binary packages.
    ``charm-strict-dependencies`` is mutually exclusive with ``charm-python-packages``.
    """

    @pydantic.field_validator("charm_entrypoint", mode="after")
    def _validate_entrypoint(cls, charm_entrypoint: str, info: pydantic.ValidationInfo) -> str:
        """Validate the entry point."""
        # the location of the project is needed
        if "source" not in info.data:
            raise ValueError(
                "cannot validate 'charm-entrypoint' because invalid 'source' configuration"
            )
        project_dirpath = pathlib.Path(info.data["source"]).resolve()

        # check that the entrypoint is inside the project
        filepath = (project_dirpath / charm_entrypoint).resolve()
        if project_dirpath not in filepath.parents:
            raise ValueError(f"charm entry point must be inside the project: {str(filepath)!r}")

        # store the entrypoint always relative to the project's path (no matter if the origin
        # was relative or absolute)
        rel_entrypoint = (project_dirpath / charm_entrypoint).relative_to(project_dirpath)
        return rel_entrypoint.as_posix()

    @pydantic.model_validator(mode="after")
    def _validate_requirements(self) -> Self:
        """Validate the specified requirement or dynamically default it.

        The default is dynamic because it's only requirements.txt if the
        file is there.
        """
        # the location of the project is needed
        if not self.source:
            raise ValueError(
                "cannot validate 'charm-requirements' because no 'source' was provided"
            )
        project_dirpath = pathlib.Path(self.source)

        # if nothing indicated, and default file is there, use it
        default_reqs_name = "requirements.txt"
        if not self.charm_requirements and (project_dirpath / default_reqs_name).is_file():
            self.charm_requirements.append(default_reqs_name)

        return self

    @pydantic.model_validator(mode="after")
    def _validate_strict_dependencies(self) -> Self:
        """Validate basic requirements if strict dependencies are enabled.

        Full validation that the requirements file contains all dependencies is done later, but
        we can fail early if the strict dependencies setting causes the charm to be invalid.
        """
        if not self.charm_strict_dependencies:
            return self

        if self.charm_python_packages:
            raise ValueError(
                "'charm-python-packages' must not be set if 'charm-strict-dependencies' is enabled"
            )

        if not self.charm_requirements:
            raise ValueError(
                "'charm-strict-dependencies' requires at least one requirements file."
            )

        invalid_binaries = set()
        for binary_package in self.charm_binary_python_packages:
            if not PACKAGE_NAME_REGEX.fullmatch(binary_package):
                invalid_binaries.add(binary_package)

        if invalid_binaries:
            raise ValueError(
                "'charm-binary-python-packages' may contain only package names allowed "
                "to be installed from binary if 'charm-strict-dependencies' is enabled. "
                f"Invalid package names: {sorted(invalid_binaries)}"
            )

        return self


class CharmPlugin(plugins.Plugin):
    """Build the charm and prepare for packing.

    The craft-parts charm plugin prepares the charm payload for packing. Common
    plugin and source keywords can be used, as well as the following optional
    plugin-specific properties:

      - ``charm-entrypoint``
        (string)
        The path to the main charm executable, relative to the charm root.

      - ``charm-binary-python-packages``
        (list of strings)
        A list of python packages to install from PyPI before installing
        requirements. Binary packages are allowed, but they may also be
        installed from sources if a package is only available in source form.

      - ``charm-python-packages``
        (list of strings)
        A list of python packages to install from PyPI before installing
        requirements. These packages will be installed from sources and built
        locally at packing time.

      - ``charm-requirements``
        (list of strings)
        List of paths to requirements files.

      - ``charm-strict-dependencies``
        (boolean)
        Whether to use legacy dependency resolution or strict dependency resolution.
        By default, legacy dependency resolution is used. If set to true, strict
        dependency resolution will be used, requiring all dependencies, including
        library dependencies, to be defined in provided requirements files.

    Extra files to be included in the charm payload must use the ``dump`` plugin.
    """

    properties_class = CharmPluginProperties

    @overrides.override
    def get_build_snaps(self) -> set[str]:
        """Return a set of required snaps to install in the build environment."""
        return set()

    def get_build_packages(self) -> set[str]:
        """Return a set of required packages to install in the build environment."""
        if platform.is_deb_based():
            return {
                "python3-dev",
                "python3-pip",
                "python3-setuptools",
                "python3-venv",
                "python3-wheel",
                "libyaml-dev",
            }
        elif platform.is_yum_based():
            try:
                os_release = os_utils.OsRelease()
                if (os_release.id(), os_release.version_id()) in (("centos", "7"), ("rhel", "7")):
                    # CentOS 7 Python 3.8 from SCL repo
                    return {
                        "autoconf",
                        "automake",
                        "gcc",
                        "gcc-c++",
                        "git",
                        "make",
                        "patch",
                        "rh-python38-python-devel",
                        "rh-python38-python-pip",
                        "rh-python38-python-setuptools",
                        "rh-python38-python-wheel",
                    }
            except (OsReleaseIdError, OsReleaseVersionIdError):
                pass

            return {
                "autoconf",
                "automake",
                "gcc",
                "gcc-c++",
                "git",
                "make",
                "patch",
                "python3-devel",
                "python3-pip",
                "python3-setuptools",
                "python3-wheel",
            }
        elif platform.is_dnf_based():
            return {
                "python3-devel",
            }
        else:
            return set()

    def get_build_environment(self) -> dict[str, str]:
        """Return a dictionary with the environment to use in the build step."""
        environment = {
            # Cryptography fails to load OpenSSL legacy provider in some circumstances.
            # Since we don't need the legacy provider, this works around that bug.
            "CRYPTOGRAPHY_OPENSSL_NO_LEGACY": "true"
        }
        os_special_paths = self._get_os_special_priority_paths()
        if os_special_paths:
            environment["PATH"] = os_special_paths + ":${PATH}"

        return environment

    def get_build_commands(self) -> list[str]:
        """Return a list of commands to run during the build step."""
        options = cast(CharmPluginProperties, self._options)

        build_env = {
            "LANG": "C.UTF-8",
            "LC_ALL": "C.UTF-8",
            # Cryptography fails to load OpenSSL legacy provider in some circumstances.
            # Since we don't need the legacy provider, this works around that bug.
            "CRYPTOGRAPHY_OPENSSL_NO_LEGACY": "true",
        }
        for key in [
            "PATH",
            "SNAP",
            "SNAP_ARCH",
            "SNAP_NAME",
            "SNAP_VERSION",
            "http_proxy",
            "https_proxy",
            "no_proxy",
        ]:
            if key in os.environ:
                build_env[key] = os.environ[key]

        os_special_paths = self._get_os_special_priority_paths()
        if os_special_paths:
            build_env["PATH"] = os_special_paths + ":" + build_env["PATH"]

        env_flags = [f"{key}={value}" for key, value in build_env.items()]

        # invoke the charm builder
        build_cmd = [
            "env",
            "-i",
            *env_flags,
            sys.executable,
            "-u",
            "-I",
            charm_builder.__file__,
            "--builddir",
            str(self._part_info.part_build_dir),
            "--installdir",
            str(self._part_info.part_install_dir),
        ]

        if options.charm_entrypoint:
            entrypoint = self._part_info.part_build_dir / options.charm_entrypoint
            build_cmd.extend(["--entrypoint", str(entrypoint)])

        if options.charm_strict_dependencies:
            build_cmd.extend(self._get_strict_dependencies_parameters())
        else:
            build_cmd.extend(self._get_legacy_dependencies_parameters())

        commands = [" ".join(shlex.quote(i) for i in build_cmd)]

        # hook a callback after the BUILD happened (to collect metrics left by charm builder)
        callbacks.register_post_step(self.post_build_callback, step_list=[Step.BUILD])

        return commands

    def _get_strict_dependencies_parameters(self) -> list[str]:
        """Get the parameters to pass to the charm builder if strict dependencies are enabled."""
        options = cast(CharmPluginProperties, self._options)
        return [
            "--strict-dependencies",
            *(f"--binary-package={package}" for package in options.charm_binary_python_packages),
            *(f"--requirement={reqs}" for reqs in options.charm_requirements),
        ]

    def _get_legacy_dependencies_parameters(self) -> list[str]:
        """Get the parameters to pass to the charm builder with strict dependencies disabled."""
        options = cast(CharmPluginProperties, self._options)
        parameters = []
        try:
            if options.charm_python_packages or options.charm_requirements:
                base_tools = ["pip", "setuptools", "wheel"]

                # remove base tools if defined in charm_python_packages
                for pkg in options.charm_python_packages:
                    pkg = re.split("[<=>]", pkg, maxsplit=1)[0].strip()
                    if pkg in base_tools:
                        base_tools.remove(pkg)

                os_release = os_utils.OsRelease()
                if (os_release.id(), os_release.version_id()) in (("centos", "7"), ("rhel", "7")):
                    # CentOS 7 compatibility, bootstrap base tools use binary packages
                    for pkg in base_tools:
                        parameters.extend(["-b", pkg])

                # build base tools from source
                for pkg in base_tools:
                    parameters.extend(["-p", pkg])
        except (OsReleaseIdError, OsReleaseVersionIdError):
            pass

        for pkg in options.charm_binary_python_packages:
            parameters.extend(["-b", pkg])

        for pkg in options.charm_python_packages:
            parameters.extend(["-p", pkg])

        for req in options.charm_requirements:
            parameters.extend(["-r", req])

        return parameters

    def post_build_callback(self, step_info):
        """Collect metrics left by charm_builder.py."""
        instrum.merge_from(env.get_charm_builder_metrics_path())

    def _get_os_special_priority_paths(self) -> str | None:
        """Return a str of PATH for special OS."""
        with suppress(OsReleaseIdError, OsReleaseVersionIdError):
            os_release = os_utils.OsRelease()
            if (os_release.id(), os_release.version_id()) in (("centos", "7"), ("rhel", "7")):
                # CentOS 7 Python 3.8 from SCL repo
                return "/opt/rh/rh-python38/root/usr/bin"

        return None



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Command for analysing a charm."""
import argparse
import json
import pathlib
from collections.abc import Container

from craft_cli import emit
from pydantic.json import pydantic_encoder

from charmcraft import errors, linters
from charmcraft.application.commands import base
from charmcraft.models import lint

OVERVIEW = """\
Analyze a charm.

Report the attributes and lint results directly in the terminal. Use
`--force` to run even those configured to be ignored.
"""


class Analyse(base.CharmcraftCommand):
    """Run analysis on a built charm."""

    name = "analyse"
    help_msg = "Analyse a charm"
    overview = OVERVIEW
    format_option = True

    def fill_parser(self, parser) -> None:
        """Add command-specific parameters."""
        super().fill_parser(parser)
        parser.add_argument(
            "--force",
            action="store_true",
            # For backwards compatibility. This command doesn't ignore suppressed linters from charmcraft.yaml.
            help=argparse.SUPPRESS,
        )
        parser.add_argument("--ignore", help="Linters to ignore (comma separated)")
        parser.add_argument("filepath", type=pathlib.Path, help="The charm to analyse")

    def run(self, parsed_args: argparse.Namespace) -> int:
        """Run the 'analyse' command."""
        if not parsed_args.filepath.exists():
            raise errors.CraftError(
                f"Charm file not found: {str(parsed_args.filepath)}",
                retcode=1,
                reportable=False,
                logpath_report=False,
            )

        ignore = parsed_args.ignore.split(",") if parsed_args.ignore else []
        if parsed_args.format:
            return self._run_formatted(parsed_args.filepath, ignore=ignore)
        return self._run_streaming(parsed_args.filepath, ignore=ignore)

    def _run_formatted(self, filepath: pathlib.Path, *, ignore=Container[str]) -> int:
        """Run the command, formatting the output into JSON or similar at the end."""
        results = list(self._services.analysis.lint_file(filepath))
        emit.message(json.dumps(results, indent=4, default=pydantic_encoder))
        return max(r.level for r in results).return_code

    def _run_streaming(self, filepath: pathlib.Path, *, ignore=Container[str]) -> int:
        """Run the command, printing linter results as we get them."""
        max_level = lint.ResultLevel.OK
        with emit.progress_bar(
            f"Linting {filepath.name}...", total=len(linters.CHECKERS)
        ) as progress:
            for result in self._services.analysis.lint_file(
                filepath, ignore=ignore, include_ignored=False
            ):
                emit.progress(str(result), permanent=True)
                max_level = max(result.level, max_level)
                progress.advance(1)

        return max_level.return_code


class Analyze(Analyse):
    """Analyse, but like a cowboy.

    US English synonym for analyse command. Hidden from the help text because
    there should be only one form of the command visible.
    """

    name = "analyze"
    hidden = True



# Copyright 2023-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for the store service."""
import platform
from typing import cast
from unittest import mock

import craft_store
import distro
import pytest
from craft_cli.pytest_plugin import RecordingEmitter
from craft_store import models
from hypothesis import given, strategies

import charmcraft
from charmcraft import application, errors, services
from charmcraft.models.project import CharmLib
from charmcraft.store import client
from tests import get_fake_revision


@pytest.fixture
def store(service_factory) -> services.StoreService:
    store = services.StoreService(app=application.APP_METADATA, services=service_factory)
    store.client = mock.Mock(spec_set=client.Client)
    store.anonymous_client = mock.Mock(spec_set=client.AnonymousClient)
    return store


@pytest.fixture(scope="module")
def reusable_store():
    store = services.StoreService(app=application.APP_METADATA, services=None)
    store.client = mock.Mock(spec_set=craft_store.StoreClient)
    return store


def test_user_agent(store):
    assert store._user_agent == f"Charmcraft/{charmcraft.__version__} ({store._ua_system_info})"


@pytest.mark.parametrize("system", ["Windows", "Macos"])
@pytest.mark.parametrize("release", ["10", "11", "12"])
@pytest.mark.parametrize("machine", ["x86_64", "arm64", "riscv64"])
@pytest.mark.parametrize("python", ["CPython", "PyPy"])
@pytest.mark.parametrize("python_version", ["3.10", "3.12"])
def test_ua_system_info_non_linux(
    monkeypatch, store, system, release, machine, python, python_version
):
    monkeypatch.setattr(platform, "system", lambda: system)
    monkeypatch.setattr(platform, "release", lambda: release)
    monkeypatch.setattr(platform, "machine", lambda: machine)
    monkeypatch.setattr(platform, "python_implementation", lambda: python)
    monkeypatch.setattr(platform, "python_version", lambda: python_version)

    assert store._ua_system_info == f"{system} {release}; {machine}; {python} {python_version}"


@pytest.mark.parametrize("machine", ["x86_64", "arm64", "riscv64"])
@pytest.mark.parametrize("python", ["CPython", "PyPy"])
@pytest.mark.parametrize("python_version", ["3.10", "3.12"])
@pytest.mark.parametrize("distro_name", ["Ubuntu", "Debian", "Something"])
@pytest.mark.parametrize("distro_version", ["1", "24.04", "version"])
def test_ua_system_info_linux(
    monkeypatch, store, machine, python, python_version, distro_name, distro_version
):
    monkeypatch.setattr(platform, "system", lambda: "Linux")
    monkeypatch.setattr(platform, "release", lambda: "6.5.0")
    monkeypatch.setattr(platform, "machine", lambda: machine)
    monkeypatch.setattr(platform, "python_implementation", lambda: python)
    monkeypatch.setattr(platform, "python_version", lambda: python_version)
    monkeypatch.setattr(distro, "name", lambda: distro_name)
    monkeypatch.setattr(distro, "version", lambda: distro_version)

    assert (
        store._ua_system_info
        == f"Linux 6.5.0; {machine}; {python} {python_version}; {distro_name} {distro_version}"
    )


def test_setup_with_error(emitter: RecordingEmitter, store):
    store.ClientClass = mock.Mock(side_effect=[craft_store.errors.NoKeyringError, "I am a store!"])

    store.setup()

    assert store.client == "I am a store!"
    emitter.assert_progress(
        "WARNING: Cannot get a keyring. Every store interaction that requires "
        "authentication will require you to log in again.",
        permanent=True,
    )


def test_get_description_default(monkeypatch, store):
    monkeypatch.setattr(platform, "node", lambda: "my-hostname")

    assert store._get_description() == "charmcraft@my-hostname"


@given(text=strategies.text())
def test_get_description_override(reusable_store, text):
    assert reusable_store._get_description(text) == text


@given(
    permissions=strategies.lists(strategies.text()),
    description=strategies.text(),
    ttl=strategies.integers(min_value=1),
    channels=strategies.lists(strategies.text()),
)
def test_login(reusable_store, permissions, description, ttl, channels):
    client = cast(mock.Mock, reusable_store.client)
    client.reset_mock(return_value=True, side_effect=True)

    reusable_store.login(
        permissions=permissions, description=description, ttl=ttl, channels=channels
    )

    client.login.assert_called_once_with(
        permissions=permissions, description=description, ttl=ttl, packages=None, channels=channels
    )


def test_login_failure(store):
    client = cast(mock.Mock, store.client)
    client.login.side_effect = craft_store.errors.CredentialsAlreadyAvailable("charmcraft", "host")

    with pytest.raises(errors.CraftError, match="Cannot login because credentials were found"):
        store.login()


def test_logout(store):
    client = cast(mock.Mock, store.client)

    store.logout()

    client.logout.assert_called_once_with()


@pytest.mark.parametrize(
    ("updates", "expected_request"),
    [
        pytest.param({}, [], id="empty"),
        pytest.param(
            {123: ["amd64", "riscv64"]},
            [
                models.CharmResourceRevisionUpdateRequest(
                    revision=123,
                    bases=[models.RequestCharmResourceBase(architectures=["amd64", "riscv64"])],
                )
            ],
        ),
        pytest.param(
            {
                123: ["amd64", "riscv64"],
                456: ["all"],
            },
            [
                models.CharmResourceRevisionUpdateRequest(
                    revision=123,
                    bases=[models.RequestCharmResourceBase(architectures=["amd64", "riscv64"])],
                ),
                models.CharmResourceRevisionUpdateRequest(
                    revision=456,
                    bases=[models.RequestCharmResourceBase(architectures=["all"])],
                ),
            ],
        ),
    ],
)
def test_set_resource_revisions_architectures_request_form(store, updates, expected_request):
    store.client.list_resource_revisions.return_value = []

    store.set_resource_revisions_architectures("my-charm", "my-file", updates)

    store.client.update_resource_revisions.assert_called_once_with(
        *expected_request,
        name="my-charm",
        resource_name="my-file",
    )


@pytest.mark.parametrize(
    ("updates", "store_response", "expected"),
    [
        ({}, [], []),
        (
            {123: ["all"]},
            [
                get_fake_revision(bases=[models.ResponseCharmResourceBase()], revision=0),
                get_fake_revision(bases=[models.ResponseCharmResourceBase()], revision=123),
            ],
            [get_fake_revision(bases=[models.ResponseCharmResourceBase()], revision=123)],
        ),
    ],
)
def test_set_resource_revisions_architectures_response_form(
    store, updates, store_response, expected
):
    store.client.list_resource_revisions.return_value = store_response

    actual = store.set_resource_revisions_architectures("my-charm", "my-file", updates)

    assert actual == expected


def test_get_credentials(monkeypatch, store):
    mock_client_class = mock.Mock(spec_set=craft_store.StoreClient)
    mock_client = mock_client_class.return_value
    monkeypatch.setattr(craft_store, "StoreClient", mock_client_class)

    store.get_credentials()

    mock_client_class.assert_called_once_with(
        application_name="charmcraft",
        base_url=mock.ANY,
        storage_base_url=mock.ANY,
        endpoints=mock.ANY,
        environment_auth=None,
        user_agent=store._user_agent,
        ephemeral=True,
    )
    mock_client.login.assert_called_once_with(
        permissions=mock.ANY,
        description=store._get_description(),
        ttl=mock.ANY,
        packages=None,
        channels=None,
    )


@pytest.mark.parametrize(
    ("libs", "expected_call"),
    [
        ([], []),
        (
            [CharmLib(lib="my_charm.my_lib", version="1")],
            [{"charm-name": "my-charm", "library-name": "my_lib", "api": 1}],
        ),
        (
            [CharmLib(lib="my_charm.my_lib", version="1.0")],
            [{"charm-name": "my-charm", "library-name": "my_lib", "api": 1, "patch": 0}],
        ),
    ],
)
def test_fetch_libraries_metadata(monkeypatch, store, libs, expected_call):

    store.get_libraries_metadata(libs)

    store.anonymous_client.fetch_libraries_metadata.assert_called_once_with(expected_call)
